# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""define things that should be visible to the user."""

# pylint: disable=wildcard-import, missing-docstring
from .track import *
from .lifetrack import *
from .backend import *
from .diff import *
from .backends import *
from .version import *

__all__ = ['Track', 'Lifetrack', 'Directory', 'GPSIES', 'MMT', 'TrackMMT', 'ServerDirectory', 'BackendDiff', 'VERSION']


def prepare_backends():
    """Initialize the attribute "supported" for all backends."""
    for key in globals().keys():
        cls = globals()[key]
        if hasattr(cls, "__mro__") and cls is not Backend:
            if cls.__mro__[-2] == Backend:
                cls._define_support()  # pylint: disable=protected-access


prepare_backends()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines some helpers."""

import os
import datetime
import logging

__all__ = ['Duration', 'repr_timespan', 'uniq', 'remove_directory', 'is_track', 'collect_tracks']


class Duration:

    """A context manager showing time information for debugging."""

    # pylint: disable=too-few-public-methods

    def __init__(self, name):
        """See class docstring."""
        self.name = name
        self.start_time = datetime.datetime.now()

    def __enter__(self):
        """See class docstring.

        Returns:
            self

        """
        return self

    def __exit__(self, exc_type, exc_value, trback):
        """See class docstring."""
        logging.debug(
            '%s in %s %s-%s',
            datetime.datetime.now() - self.start_time,
            self.name, self.start_time, datetime.datetime.now())


def repr_timespan(start, end) ->str:
    """return a string representing the timespan.

    Returns:
        a string like #h#m

    """
    duration = end - start
    hours = duration.seconds // 3600
    minutes = (duration.seconds - hours * 3600) // 60
    hours += duration.days * 24
    return '{}:{:02}'.format(hours, minutes)


def uniq(lst):
    """return lst with unique elements."""
    seen = []
    for _ in lst:
        if _ not in seen:
            seen.append(_)
            yield _


def remove_directory(path):
    """If this fails, show directory content."""
    try:
        os.rmdir(path)
    except OSError as exc:
        logging.error('rmdir: errno: %s cannot remove directory: %s', exc.errno, path)
        if os.path.exists(path):
            for _ in os.listdir(path):
                logging.error('  dir still has %s', _)


def is_track(value):
    """Return True or False without looking at the type, so we do not need to import Track."""
    return hasattr(value, 'id_in_backend')


def collect_tracks(sources):
    """A copied list with tracks combined from all sources, to be used in 'for'-loops.

    Returns:
        A list of tracks

    """
    if is_track(sources):
        return [sources]
    result = list()
    for source in sources:
        logging.debug('collecting tracks from %s', source.identifier() or '.')
        if is_track(source):
            result.append(source)
        else:
            result.extend(source)
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.Lifetrack`."""

from .track import Track

__all__ = ['Lifetrack']


class LifetrackTarget:

    """A single target of a lifetracking instance."""

    def __init__(self, backend):
        """See class docstring."""
        self.backend = backend
        self.track = Track()
        self.__started = False

    def update(self, points) ->str:
        """Update lifetrack into a specific track.

        Returns:
            the new id_in_backend

        """

        # pylint: disable=protected-access

        if not points:
            if not self.__started:
                raise Exception('Lifetrack needs initial points')
            else:
                raise Exception('Lifetrack.update needs points')
        if 'Lifetracking' in self.track.title:
            with self.track._decouple():
                self.track.title = ':'.join(self.track.title.split(':')[1:])
        old_title = self.track.title
        new_ident = None
        try:
            if 'Lifetracking' not in self.track.title:
                with self.track._decouple():
                    self.track.title = (
                        'Lifetracking continues:'
                        if self.__started else 'Lifetracking starts:') + ' ' + self.track.title
            if 'lifetrack' in self.backend.supported:
                if not self.__started:
                    new_ident = self.backend._lifetrack_start(self.track, self._prepare_points(points))
                    with self.track._decouple():
                        self.track.id_in_backend = new_ident
                else:
                    self.backend._lifetrack_update(self.track, self._prepare_points(points))
            else:
                self.track.add_points(points)
                if not self.__started:
                    self.track = self.backend.add(self.track)
                    new_ident = self.track.id_in_backend
                    assert new_ident
                assert self.track.id_in_backend
            return new_ident
        finally:
            with self.track._decouple():
                self.track.title = old_title
            self.__started = True

    def end(self):
        """End lifetracking for a specific backend."""
        if not self.__started:
            raise Exception('Lifetrack not yet started')
        if 'lifetrack_end' in self.backend.supported:
            with self.track._decouple():  # pylint: disable=protected-access
                self.track.title = 'Lifetracking ends: ' + self.track.title
            self.backend._lifetrack_end(self.track)  # pylint: disable=protected-access

    def _prepare_points(self, points):
        """Round points.

        Returns:
            The rounded points

        """
        result = list(points)[:]
        self.track._round_points(result)  # pylint: disable=protected-access
        return result


class Lifetrack:

    """Life tracking. The data will be forwarded to all given backends.

    Args:
        target_tracks: Those tracks will receive the lifetracking data.
            This may be a list(Track) or a single track.

    """

    # pylint: disable=protected-access

    def __init__(self, target_backends):
        """See class docstring."""
        if isinstance(target_backends, list):
            _ = target_backends
        else:
            _ = [target_backends]

        self.targets = [LifetrackTarget(x) for x in _]
        self.id_in_server = None

    def update(self, points):
        """Start or update lifetrack.

        If the backend does not support lifetrack, this just saves the track in the backend.

        Args:
            points(list): New points

        """
        for _ in self.targets:
            id_in_server = _.update(points)
            assert _.track.id_in_backend, '{} in {} got no id_in_backend'.format(_.track, _.backend)
            if self.id_in_server is None:
                self.id_in_server = id_in_server

    def end(self):
        """End lifetrack.

        If the backend does not support lifetrack, this does nothing."""
        for _ in self.targets:
            _.end()

    def set_title(self, title):
        """Set title for all targets."""
        for _ in self.targets:
            with _.track._decouple():
                _.track.title = title

    def set_category(self, category):
        """Set category for all targets."""
        for _ in self.targets:
            with _.track._decouple():
                _.track.category = category

    def set_public(self, public):
        """Set public for all targets."""
        for _ in self.targets:
            with _.track._decouple():
                _.track.public = public



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
implements test classes for Track.

They only use backend Directory, so there is no network traffic involved
(unless Directory is a network file system, of course).

"""

import os
import io
import filecmp
import tempfile
import datetime

from .basic import BasicTest
from ... import Track
from .. import Directory
from ...util import repr_timespan

# pylint: disable=attribute-defined-outside-init


class TrackTests(BasicTest):

    """track tests."""

    def test_init(self):
        """test initialisation."""
        track = Track()
        self.assertFalse(track.public)
        with Directory(cleanup=True) as backend:
            track = Track()
            track._set_backend(backend)  # pylint: disable=protected-access
            self.assertEqual(len(backend), 0)
            backend.add(track)
            self.assertEqual(len(backend), 1)

        with self.temp_backend(Directory, count=2) as backend:
            backend.add(Track())
            self.assertEqual(len(backend), 3)

        test_url = tempfile.mkdtemp(prefix=Directory.prefix)
        self.assertTrue(os.path.exists(test_url))
        os.rmdir(test_url)
        self.assertFalse(os.path.exists(test_url))
        try:
            with Directory(url=test_url, cleanup=True):
                self.assertTrue(os.path.exists(test_url))
        finally:
            os.rmdir(test_url)

    def test_track_list(self):
        """test list of tracks."""
        with Directory(cleanup=True) as directory:
            self.assertEqual(len(directory), 0)
            track1 = Track()
            directory.add(track1)
            self.assertIn(track1, directory)
            self.assertIsNotNone(track1.id_in_backend)
            track1.description = 'x'
            self.assertIsNotNone(track1.id_in_backend)

    def test_clone(self):
        """True if the clone is identical."""
        track1 = self.create_test_track()
        track2 = track1.clone()
        self.assertEqualTracks(track1, track2)
        count1 = track1.gpx.get_track_points_no()
        del track1.gpx.tracks[0].segments[0].points[0]
        self.assertEqual(count1, track1.gpx.get_track_points_no() + 1)
        self.assertNotEqualTracks(track1, track2)
        track2 = track1.clone()
        track2.gpx.tracks[-1].segments[-1].points[-1].latitude = 5
        self.assertNotEqualTracks(track1, track2)
        track2 = track1.clone()
        track2.gpx.tracks[-1].segments[-1].points[-1].longitude = 5
        self.assertNotEqual(track1, track2)
        track2 = track1.clone()
        last_point2 = track2.gpx.tracks[-1].segments[-1].points[-1]
        last_point2.elevation = 500000
        self.assertEqual(last_point2.elevation, 500000)
        # here assertNotEqualTracks is wrong because keys() are still identical
        self.assertTrue(track1.points_equal(track2))
        track1.gpx.tracks.clear()
        track2.gpx.tracks.clear()
        self.assertEqualTracks(track1, track2)

    def test_no_category(self):
        """category must return default value if not present in gpx.keywords."""
        category_default = Track.legal_categories[0]
        track = Track()
        self.assertEqual(track.category, category_default)
        track.category = None
        self.assertEqual(track.category, category_default)
        with self.assertRaises(Exception):
            track.category = 'illegal value'
        self.assertEqual(track.category, category_default)
        with self.assertRaises(Exception):
            track.add_keywords('Category:illegal value')
        self.assertEqual(track.category, category_default)

    def test_duplicate_category(self):
        """try to add two categories to Track."""
        category_other = Track.legal_categories[5]
        track = Track()
        track.category = category_other
        with self.assertRaises(Exception):
            track.add_keywords('Category:{}'.format(category_other))

    def test_remove_category(self):
        """remove category from Track."""
        category_default = Track.legal_categories[0]
        category_other = Track.legal_categories[5]
        track = Track()
        track.category = category_other
        self.assertEqual(track.category, category_other)
        track.category = None
        self.assertEqual(track.category, category_default)

    def test_no_public(self):
        """public must return False if not present in gpx.keywords."""
        track = Track()
        self.assertFalse(track.public)

    def test_duplicate_public(self):
        """try to set public via its property and additionally with add_keywords."""
        track = Track()
        track.public = True
        self.assertTrue(track.public)
        with self.assertRaises(Exception):
            track.add_keywords('Status:public')

    def test_remove_public(self):
        """remove and add public from Track using remove_keywords and add_keywords."""
        track = Track()
        track.public = True
        with self.assertRaises(Exception):
            track.remove_keywords('Status:public')
        self.assertTrue(track.public)
        with self.assertRaises(Exception):
            track.add_keywords('Status:public')
        self.assertTrue(track.public)

    def test_last_time(self):
        """Track.last_time."""
        track = self.create_test_track()
        gpx_last_time = track.gpx.tracks[-1].segments[-1].points[-1].time
        self.assertEqual(track.last_time, gpx_last_time)

    def test_one_line_per_trkpt(self):
        """One line per trackpoint."""
        track = self.create_test_track()
        xml = track.to_xml()
        self.assertNotIn('<link ></link>', xml)
        lines = xml.split('\n')
        self.logger.debug('xml is:%s', xml)
        start_lines = {x for x in lines if x.strip().startswith('<trkpt')}
        end_lines = {x for x in lines if x.strip().endswith('</trkpt>')}
        have_points = track.gpx.get_track_points_no()
        self.assertEqual(len(start_lines), have_points)
        self.assertEqual(len(end_lines), have_points)
        self.assertEqual(start_lines, end_lines)

    def test_parse(self):
        """check for Track parsing xml correctly."""
        track = self.create_test_track()
        track.keywords = ['Here are some keywords']
        xml = track.to_xml()
        track2 = Track()
        track2.parse(None)
        track2.parse('')
        track2.parse(xml)
        self.assertEqualTracks(track, track2)
        self.assertEqual(track.keywords, track2.keywords)
        track2 = Track()
        track2.parse(io.StringIO(xml))
        self.assertEqualTracks(track, track2)

    def test_combine(self):
        """combine values in track with newly parsed."""
        # first, does it overwrite?
        track = self.create_test_track()
        xml = track.to_xml()
        if track.category == 'Cycling':
            other_category = 'Running'
        else:
            other_category = 'Cycling'

        track2 = Track()
        track2.title = 'Title2'
        track2.description = 'Description2'
        track2.category = other_category
        track2.public = True
        track2.parse(xml)
        self.assertEqual(track2.title, track.title)
        self.assertEqual(track2.description, track.description)
        self.assertEqual(track2.category, track.category)
        self.assertTrue(track2.public)
        self.assertEqual(track2.keywords, list())

        track.public = True
        xml = track2.to_xml()
        self.assertIn('Status:public', xml)
        track2 = Track()
        track2.category = Track.legal_categories[3]
        track2.public = False
        track2.parse(xml)
        self.assertTrue(track2.public)

        # second, does it keep old values if there are no new values?
        track = self.create_test_track()
        track.title = ''
        track.description = ''
        xml = track.to_xml()
        if track.category == 'Cycling':
            other_category = 'Running'
        else:
            other_category = 'Cycling'

        track2 = Track()
        track2.title = 'Title2'
        track2.description = 'Description2'
        track2.parse(xml)
        self.assertEqual(track2.title, 'Title2')
        self.assertEqual(track2.description, 'Description2')

    def test_save_dir(self):
        """Correct files?."""
        with Directory(cleanup=True) as directory:
            os.chmod(directory.url, 0o555)
            track = self.create_test_track()
            if os.getuid() == 0:
                # for root, this works even with 555
                directory.add(track)
                self.assertIsNotNone(track.backend)
            else:
                with self.assertRaises(OSError):
                    directory.add(track)
                self.assertIsNone(track.backend)
                os.chmod(directory.url, 0o755)
                directory.add(track)
            self.assertIsNotNone(track.backend)

    def test_save(self):
        """save locally."""
        with Directory(cleanup=True) as directory:
            dir2 = self.clone_backend(directory)
            try:
                track = self.create_test_track()
                directory.add(track)
                self.assertEqual(len(directory), 1)
                aclone = track.clone()
                self.assertEqualTracks(track, aclone)

                self.assertEqual(len(dir2), 1)

                track2 = track.clone()
                self.assertEqualTracks(track, track2)
                directory.add(track2)
                self.assertEqual(len(directory), 2)
                dir2.add(track2)
                self.assertEqual(len(dir2), 2)

                track2_copy = dir2.add(track2.clone())
                self.assertEqualTracks(track, track2_copy)
                self.assertEqualTracks(track2, track2_copy)
                self.assertIs(track.backend, directory)
                self.assertIs(track2.backend, directory)
                self.assertIs(track2_copy.backend, dir2)
                self.assertEqual(len(directory), 2)
                self.assertEqual(len(dir2), 3)
                directory.scan()  # we changed it through dir2
                self.assertEqual(len(directory), 4)
                dir2.scan()
                self.assertEqual(len(directory), 4)
                title = 'whatevertitle'
                for _ in dir2:
                    _.id_in_backend = title
                trunk = os.path.join(directory.url, title)
                expected_names = [trunk + x + '.gpx' for x in ('.1', '.2', '.3', '')]
                files = sorted(
                    os.path.join(directory.url, x)
                    for x in os.listdir(directory.url) if x.endswith('.gpx'))
                self.assertEqual(files, expected_names)
                self.assertEqual(len(dir2), 4)
                directory.scan()
                dir2.merge(directory, remove=True)
                self.assertEqual(len(dir2), 1)
                filecmp.clear_cache()
            finally:
                dir2.destroy()

    def test_add_points(self):
        """test Track.add_points."""
        point_count = 11
        track = Track()
        points = self._random_points(count=point_count)
        track.add_points(points)
        self.assertEqual(track.gpx.get_track_points_no(), point_count)
        with self.assertRaises(Exception):
            track.add_points(points)
        track.add_points(points[:-1])
        self.assertEqual(track.gpx.get_track_points_no(), point_count * 2 - 1)

    def test_points_equal(self):
        """test Track.points_equal."""
        for _ in range(100):
            points = self._random_points(count=7)
            track1 = Track()
            track1.add_points(points)
            track2 = track1.clone()
            points2 = list(track2.points())  # those are cloned points
            self.assertTrue(track1.points_equal(track2))
            track2.gpx.tracks.clear()
            track2.add_points(points2[:5])
            self.assertFalse(track1.points_equal(track2))
            track2.add_points(points2[5:])
            self.assertTrue(track1.points_equal(track2))

            old_long = track2.gpx.tracks[-1].segments[-1].points[-2].longitude
            track2.gpx.tracks[-1].segments[-1].points[-2].longitude += 1
            self.assertFalse(track1.points_equal(track2))
            track2.gpx.tracks[-1].segments[-1].points[-2].longitude = old_long
            self.assertTrue(track1.points_equal(track2))

            old_lat = track2.gpx.tracks[-1].segments[-1].points[-2].latitude
            track2.gpx.tracks[-1].segments[-1].points[-2].latitude += 1
            self.assertFalse(track1.points_equal(track2))
            track2.gpx.tracks[-1].segments[-1].points[-2].latitude = old_lat
            self.assertTrue(track1.points_equal(track2))

            track2.gpx.tracks[-1].segments[-1].points[-2].elevation += 1
            self.assertTrue(track1.points_equal(track2))
            track2.gpx.tracks[-1].segments[-1].points[-2].elevation -= 1

            old_long = track2.gpx.tracks[-1].segments[-1].points[-1].longitude
            track2.gpx.tracks[-1].segments[-1].points[-1].longitude += 1
            self.assertFalse(track1.points_equal(track2))
            a1_points = list(track1.points())
            a2_points = list(track2.points())
            a1_first = a1_points[0]
            a1_last = a1_points[-1]
            a2_first = a2_points[0]
            a2_last = a2_points[-1]
            self.assertNotEqual(
                track1.angle(), track2.angle(),
                'a1.first:{} a1.last:{} a2.first:{} a2.last:{}'.format(
                    a1_first, a1_last, a2_first, a2_last))
            track2.gpx.tracks[-1].segments[-1].points[-1].longitude = old_long
            self.assertTrue(track1.points_equal(track2))

    def test_str(self):
        """test __str__."""
        track = Track()
        self.assertNotIn('id:', str(track))
        with Directory(cleanup=True) as directory:
            track = Track()
            track.title = 'Title'
            track.category = 'Running'
            track.add_points(self._random_points(10))
            self.assertIn('Title', str(track))
            self.assertIn('public' if track.public else 'private', str(track))
            self.assertIn('Running', str(track))
            self.assertIn(repr_timespan(track.time, track.last_time), str(track))
            self.assertTrue(str(track).startswith('Track('))
            self.assertTrue(str(track).endswith(')'))
            track.add_points(self._random_points(count=5))
            self.assertIn(' 15 points', str(track))
            directory.add(track)
            self.assertIn('id:', str(track))

            # str(track) must not fully load it
            clone = self.clone_backend(directory)
            self.assertIn(' 0 points', str(clone[0]))
            self.assertEqual(clone[0].gpx.get_track_points_no(), 15)
            self.assertIn(' 15 points', str(clone[0]))

    def test_angle(self):
        """test Track.angle."""
        track1 = Track()
        track1.add_points(list())
        self.assertEqual(len(track1.gpx.tracks), 0)
        self.assertEqual(track1.angle(), 0)
        track1.add_points(self._random_points(1))
        del track1.gpx.tracks[0].segments[0]
        self.assertEqual(track1.angle(), 0)
        for _ in range(1000):
            track1 = Track()
            track1.add_points(self._random_points(2))
            angle = track1.angle()
            self.assertLess(angle, 360.001)
            self.assertGreater(angle, -0.001)

        track1 = Track()
        track1.add_points(self._random_points(2))
        first_point = None
        for point in track1.points():
            if first_point is None:
                first_point = point
            else:
                point.latitude = first_point.latitude
                point.longitude = first_point.longitude
        self.assertEqual(track1.angle(), 0)

    def test_key(self):
        """Track.key()."""
        title = 'This is a niße title'
        description = title + ' NOT - it is the description'
        category = Track.legal_categories[3]
        public = True
        points = self._random_points(10)
        track = Track()
        track.title = title
        track.description = description
        track.category = category
        track.public = public
        track.add_points(points)
        key = track.key()
        self.assertIn('title:{}'.format(title), key)
        self.assertIn('description:{}'.format(description), key)
        self.assertIn('category:{}'.format(category), key)
        self.assertIn('public:True', key)
        self.assertIn('last_time:{}'.format(track.last_time), key)
        self.assertIn('angle:{}'.format(track.angle()), key)
        self.assertIn('points:{}'.format(track.gpx.get_track_points_no()), key)

    def test_symlinks(self):
        """Directory symlinks."""
        with Directory(cleanup=True) as directory:
            source = os.path.join(directory.url, 'deadlink')
            target = 'deadtarget'
            target_path = os.path.join(directory.url, target)
            with open(target_path, 'w') as target_file:
                target_file.write(' ')
            os.symlink('deadtarget', source)
            os.remove(target_path)
            directory.scan()  # this loads symlinks. It removes the dead link.
            self.assertFalse(os.path.exists(source))

    def test_fs_encoding(self):
        """fs_encoding."""
        with Directory(cleanup=True) as directory:
            track = Track()
            directory.add(track)
            org_ident = track.id_in_backend
            track.title = 'TITLE'
            self.assertEqual(track.id_in_backend, org_ident)
            self.assertEqual(track.title, 'TITLE')
            track.title = 'Tätel'
            self.assertEqual(track.title, 'Tätel')
            for title in ('a/b', '//', 'Ä/Ü', '...'):
                track.title = title
                self.assertEqual(track.title, title)
                self.assertNotEqual(track.id_in_backend, title)
                track.id_in_backend = track.title
                self.assertEqual(track.id_in_backend, title.replace('/', '_'))
            prev_encoding = directory.fs_encoding
            directory.fs_encoding = 'whatever'
            try:
                with self.assertRaises(Exception):
                    track.id_in_backend = 'TITLE'
            finally:
                directory.fs_encoding = prev_encoding

    def test_local_keywords(self):
        """Some keyword tests. More see in test_backends."""
        # Category: and Status: are special
        gpx = self._get_gpx_from_test_file('test')
        gpx.keywords = 'Category:Cycling, Status:public'
        track = Track(gpx=gpx)
        self.assertEqual(track.keywords, list())

        # : is legal within a keyword
        gpx.keywords = 'Hello:Dolly'
        track = Track(gpx=gpx)
        self.assertEqual(track.keywords, list(['Hello:Dolly']))

        # keywords are sorted
        gpx.keywords = 'Hello,Dolly'
        track = Track(gpx=gpx)
        self.assertEqual(track.keywords, list(['Dolly', 'Hello']))

        # no comma within a keyword
        with self.assertRaises(Exception):
            track.add_keywords(['Bye,Sam'])

        # keywords as string
        track.add_keywords('Bye,Sam')
        self.assertEqual(track.keywords, ['Bye', 'Dolly', 'Hello', 'Sam'])

    def test_keyword_args(self):
        """'Track.keywords' must accept any variant of iterable."""
        track = Track()
        test_tracks = list(sorted(['a', self.unicode_string2]))
        track.keywords = set(test_tracks)
        self.assertEqual(track.keywords, test_tracks)
        track.keywords = reversed(test_tracks)
        self.assertEqual(track.keywords, test_tracks)
        track.add_keywords(test_tracks[0])
        self.assertEqual(track.keywords, test_tracks)
        track.keywords = test_tracks * 2
        self.assertEqual(track.keywords, test_tracks)

    def test_id(self):
        """id_in_backend must be str."""
        with Directory(cleanup=True) as directory:
            track = Track()
            with self.assertRaises(Exception):
                directory.add(track).id_in_backend = 56
            self.assertEqual(len(directory), 1)
            with self.assertRaises(ValueError):
                directory.add(track)
            directory.add(track.clone())
            self.assertEqual(len(directory), 2)

    def test_in(self):
        """x in backend."""
        with Directory(cleanup=True) as directory:
            track = Track()
            directory.add(track).id_in_backend = '56'
            self.assertEqual(track.id_in_backend, '56')
            self.assertIn(track, directory)
            self.assertIn(track.id_in_backend, directory)
            directory.remove_all()
            self.assertNotIn(track, directory)
            self.assertNotIn(track.id_in_backend, directory)

    def test_getitem(self):
        """backend[idx]."""
        with Directory(cleanup=True) as directory:
            directory.scan(now=True)
            track = Track()
            directory.add(track).id_in_backend = '56'
            self.assertIs(directory[0], track)
            self.assertIs(directory[track], track)
            self.assertIs(directory['56'], track)
            directory.remove_all()
            with self.assertRaises(IndexError):
                directory[0]  # pylint: disable=pointless-statement

    def test_adjust_time(self):
        """adjust_time()."""
        track = self.create_test_track()
        first_wp_time = track.gpx.waypoints[0].time
        first_trkpt_time = next(track.points()).time
        seconds10 = datetime.timedelta(seconds=10)
        track.adjust_time(seconds10)
        self.assertEqual(track.gpx.waypoints[0].time, first_wp_time + seconds10)
        self.assertEqual(next(track.points()).time, first_trkpt_time + seconds10)

    def test_overlapping_times(self):
        """Track.overlapping_times(tracks)."""
        now = datetime.datetime.now()
        track1 = self.create_test_track(start_time=now)
        seconds10 = datetime.timedelta(seconds=10)
        track2 = self.create_test_track(start_time=track1.last_time - seconds10)
        track3 = self.create_test_track(start_time=track1.last_time)
        self.assertEqual(track1.last_time - seconds10, track2.time)
        group1 = list([track1, track2, track3])
        track4 = self.create_test_track(start_time=track3.last_time + seconds10)
        group2 = list([track4, track4])
        self.assertEqual(list(Track.overlapping_times(group1 + group2)), list([group1, group2]))
        group2 = list([track4])
        self.assertEqual(list(Track.overlapping_times(group1 + group2)), list([group1]))

    def test_header_changes(self):
        """Only change things in _header_data. Assert that the full gpx is loaded before saving."""
        with self.temp_backend(Directory, count=1) as backend:
            backend2 = self.clone_backend(backend)
            backend2[0].description = 'test'
            self.assertTrackFileContains(backend2[0], '<trk>')
        with self.temp_backend(Directory, count=1) as backend:
            backend2 = Directory(url=backend.url)
            backend2[0].title = 'test title'
            self.assertTrackFileContains(backend2[0], '<trk>')
        with self.temp_backend(Directory, count=1) as backend:
            backend2 = Directory(url=backend.url)
            backend2[0].category = 'Mountain biking'
            self.assertTrackFileContains(backend2[0], '<trk>')

    def test_remove_track(self):
        """If a backend has several identical tracks, make sure we remove the right one."""
        with self.temp_backend(Directory, count=1) as backend:
            track = backend[0]
            track_id = track.id_in_backend
            track2 = track.clone()
            backend.add(track2)
            backend.remove(track2)
            self.assertEqual(backend[0].id_in_backend, track_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.Backend`."""

import os
import datetime
from inspect import getmembers, isfunction
import dis
from contextlib import contextmanager
from collections import defaultdict
import logging

from .auth import Authenticate
from .track import Track
from .util import collect_tracks

__all__ = ['Backend']


class Backend:

    """A place where tracks live. Something like the filesystem or http://mapmytracks.com.

    A Backend should hold only tracks for one person, and they
    should not overlap in time. This is not enforced but sometimes
    behaviour is undefined if you ignore this.

    A Backend be used as a context manager. Upon termination, all tracks
    may be removed automatically by setting cleanup=True. Some concrete
    implementations may also remove the backend itself.

    A Backend allows indexing by normal int index, by :class:`Track <gpxity.Track>`
    and by :attr:`Track.id_in_backend <gpxity.Track.id_in_backend>`.
    :literal:`if 'ident' in backend` is possible.
    len(backend) shows the number of tracks. Please note that Code
    like :literal:`if backend:` may not behave as expected. This will be False if the backend
    has no track. If that is not what you want, consider :literal:`if backend is not None`

    The backend will automatically synchronize. So something like :literal:`len(Backend())` will work.
    However, some other Backend pointing to the same storage or even a different process
    might change things. If you want to cope with that, use :meth:`scan`.

    Not all backends support all methods. The unsupported methods
    will raise NotImplementedError. As a convenience every backend
    has a list **supported** to be used like :literal:`if 'track' in backend.supported:`
    where `track` is the name of the method.

    Backends support no locking. If others modify a backend concurrently, you may
    get surprises. It is up to you to handle those.

    Some backends may use cookies.

    Args:
        url (str): Initial value for :attr:`url`
        auth (tuple(str, str)): (username, password). Alternatively you can pass the username as a single string.
            This will lookup the password from :class:`Authenticate <gpxity.auth.Authenticate>`.
        cleanup (bool): If true, :meth:`destroy` will remove all tracks.

    Attributes:
        supported (set(str)): The names of supported methods. Creating the first instance of
            the backend initializes this. Only methods which may not be supported are mentioned here.
            If a particular value write_* like write_public does not exist, the entire track is written instead
            which normally results in a new ident for the track.
        full_support (set(str)): All possible values for the supported attribute.
        url (str): the address. May be a real URL or a directory, depending on the backend implementation.
            Every implementation may define its own default for url.
        timeout: If None, there are no timeouts: Gpxity waits forever. For legal values
            see http://docs.python-requests.org/en/master/user/advanced/#timeouts
        config: A dict with all entries in auth.cfg for this backend


    """

    # pylint: disable=too-many-instance-attributes

    class NoMatch(Exception):
        """Is raised if a track is expected to pass the match filter but does not"""

    class BackendException(Exception):
        """Is raised for general backend exceptions, especially error messages from a remote server"""

    supported = None

    skip_test = False

    _legal_categories = None  # Override in the backends

    default_url = None  # Override in the backends

    full_support = (
        'scan', 'remove', 'lifetrack', 'lifetrack_end', 'get_time', 'write', 'write_title', 'write_public',
        'write_category', 'write_gpx', 'write_description', 'write_add_keywords', 'write_remove_keywords')

    # It is important that we have only one global session per identifier()
    # because gpsies.com seems to have several servers and their
    # synchronization is sometimes slower than expected. See
    # cookie "SERVERID".
    _session = dict()

    def __init__(self, url: str = None, auth=None, cleanup: bool = False, timeout=None):
        """See class docstring."""
        if self.is_disabled():
            raise Backend.BackendException('class {} is disabled'.format(self.__class__.__name__))
        self._decoupled = False
        super(Backend, self).__init__()
        self.__tracks = list()
        self._tracks_fully_listed = False
        self.url = url or ''
        if isinstance(auth, str):
            _ = Authenticate(self.__class__, auth)
            self.auth = _.auth
            self.config = _.section
            if _.url:
                self.url = _.url
        else:
            self.config = dict()
            self.auth = auth
        if self.url and not self.url.endswith('/'):
            self.url += '/'
        self._cleanup = cleanup
        self.__match = None
        self.logger = logging.getLogger(self.identifier())
        self.timeout = timeout
        self._current_track = None

    def identifier(self, track=None) ->str:
        """Used for formatting strings. A unique identifier for every physical backend.

        Two Backend() instances pointing to the same physical backend have the same identifier.

        Args:
            track: If given, add it to the identifier.

        Returns:
            A unique identifier

        """
        result = '{}:{}{}/'.format(
            self.__class__.__name__.lower(),
            '' if self.url == self.default_url else self.url,
            self.auth[0] if self.auth and self.auth[0] else '')
        if track:
            result += track.id_in_backend
        return result

    @property
    def legal_categories(self):
        """
        A list with all legal categories.

        Returns: list(str)
            all legal values for this backend

        """
        raise NotImplementedError

    @contextmanager
    def _decouple(self):
        """Context manager: disable automic synchronization with the backend.

        In that state, automatic writes of changes into
        the backend are disabled, and if you access attributes which
        would normally trigger a full load from the backend, they will not.
        Use this to avoid recursions.

        You should never need this unless you implement a new backend.

        """
        prev_decoupled = self._decoupled
        self._decoupled = True
        try:
            yield
        finally:
            self._decoupled = prev_decoupled

    @classmethod
    def _is_implemented(cls, method) ->bool:
        """False if the first instruction in method raises NotImplementedError or if the method does nothing.

        Returns:
            True if method is implemented

        """
        first_instruction = next(dis.get_instructions(method.__code__))
        return first_instruction is not None and first_instruction.argval != 'NotImplementedError'

    @classmethod
    def _define_support(cls):
        """If the first thing a method does is raising NotImplementedError, it is marked as unsupported."""
        support_mappings = {
            # map internal names to more user friendly ones. See doc for
            # Backend.supported.
            '_lifetrack_end': 'lifetrack_end',
            '_lifetrack_start': 'lifetrack',
            '_remove_ident': 'remove',
            '_write_all': 'write',
            '_yield_tracks': 'scan',
            'get_time': 'get_time'}
        cls.supported = set()
        for name, method in getmembers(cls, isfunction):
            if name in support_mappings:
                if cls._is_implemented(method):
                    cls.supported.add(support_mappings[name])
            elif name.startswith('_write_') and name != '_write_attribute':
                if cls._is_implemented(method):
                    cls.supported.add(name[1:])

    @property
    def match(self):
        """Filter tracks.

        A function with one argument returning None or str. The backend will call
            this with every track and ignore tracks where match does not return None.
            The returned str should explain why the track does not match.

            If you change a track such that it does not match anymore, the exception
            NoMatch will be raised and the match stays unchanged.

        """
        return self.__match

    @match.setter
    def match(self, value):
        """see match.getter."""
        old_match = self.__match
        self.__match = value
        try:
            self.scan()
        except self.NoMatch:
            self.__match = old_match
            raise

    def decode_category(self, value: str) ->str:
        """Translate the value from the backend into internal one."""
        raise NotImplementedError

    def encode_category(self, value: str) ->str:
        """Translate internal value into the backend specific value."""
        raise NotImplementedError

    @staticmethod
    def _encode_keyword(value: str) ->str:
        """Replicate the translation the backend does. MMT for example capitalizes all words.

        Returns:
            the encoded keyword

        """
        return value

    def get_time(self) ->datetime.datetime:
        """get time from the server where backend is located as a Linux timestamp.
        A backend implementation does not have to support this."""
        raise NotImplementedError()

    def _change_id(self, track, new_ident: str):
        """Change the id in the backend."""
        raise NotImplementedError

    def scan(self, now: bool = False) ->None:
        """Enforce a reload of the list of all tracks in the backend.

        This will be delayed until the list is actually needed again.

        If this finds an unsaved track not matching the current match
        function, an exception is thrown.
        Saved Tracks not matching the current match will no be loaded.

        Args:
            now: If True, do not delay scanning.

        """
        self._tracks_fully_listed = False
        if now:
            self._scan()

    def _scan(self) ->None:
        """load the list of all tracks in the backend if not yet done.
        Enforce this by calling :meth:`scan` first.
        """
        if not self._tracks_fully_listed and not self._decoupled:
            self._tracks_fully_listed = True
            unsaved = [x for x in self.__tracks if x.id_in_backend is None]
            if self.__match is not None:
                for track in unsaved:
                    self.matches(track, 'scan')
            self.__tracks = unsaved
            match_function = self.__match
            self.__match = None
            try:
                # _yield_tracks should return ALL tracks, match will be
                # applied in a second loop. This way the Backend implementations
                # do not have to worry about the match code.
                with self._decouple():
                    list(self._yield_tracks())
            finally:
                self.__match = match_function
            if self.__match is not None:
                self.__tracks = [x for x in self.__tracks if self.matches(x)]

    def _found_track(self, ident: str):
        """Create an empty track for ident and inserts it into this backend.

        Returns:
            the new track

        """
        result = Track()
        with self._decouple():
            result._set_backend(self)  # pylint: disable=protected-access
            result.id_in_backend = ident
        self.__append(result)
        return result

    def _yield_tracks(self):
        """A generator for all tracks. It yields the next found and appends it to tracks.

        The tracks will not be loaded if possible.

        Yields:
            the next track

        """
        raise NotImplementedError()

    def _read_all_decoupled(self, track) ->None:
        """Decouple and calls the backend specific _read_all."""
        with self._decouple():
            self._read_all(track)

    def _read_all(self, track) ->None:
        """fill the track with all its data from source."""
        raise NotImplementedError()

    def matches(self, track, exc_prefix: str = None):
        """match track against the current match function.

        Args:
            exc_prefix: If not None, use it for the beginning of an exception message.
                If None, never raise an exception

        Returns:
            True for match

        """
        if self.__match is None:
            return True
        match_error = self.__match(track)
        if match_error and exc_prefix:
            raise Backend.NoMatch('{}: {} does not match: {}'.format(exc_prefix, track, match_error))
        return match_error is None

    def _needs_full_save(self, changes) ->bool:
        """Do we have to rewrite the entire track?.

        Returns:
            True if we must save fully

        """
        for change in changes:
            if change == 'all':
                return True
            write_name = 'write_{}'.format(change.split(':')[0])
            if write_name not in self.supported:
                return True
        return False

    def add(self, track):
        """
        Add a track to this backend.

        We do not check if it already exists in this backend. No track
        already existing in this backend will be overwritten, the id_in_backend
        of track will be deduplicated if needed. This is currently only needed
        for Directory. Note that some backends reject a track if it is very
        similar to an existing track even if it belongs to some other user.

        If the track object is already in the list of tracks, raise ValueError.

        If the track does not pass the current match function, raise an exception.

        Args:
            track (~gpxity.Track): The track we want to save in this backend.

        Returns:
            ~gpxity.Track: The saved track. If the original track lives in a different
            backend, a new track living in this backend will be created
            and returned.

        """
        if self._decoupled:
            raise Exception('A backend cannot save() while being decoupled. This is probably a bug in gpxity.')
        self._current_track = track
        self.matches(track, 'add')
        if track.backend is not self and track.backend is not None:
            new_track = track.clone()
        else:
            if any(x is track for x in self.__tracks):
                raise ValueError('Already in list: Track {} with id={}'.format(track, id(track)))
            new_track = track
        with self._decouple():
            new_track._set_backend(self)  # pylint: disable=protected-access
            if track.keywords:
                _ = (x.strip() for x in track.keywords)
                track.gpx.keywords = ', '.join(self._encode_keyword(x) for x in _)
        try:
            with self._decouple():
                self._write_all(new_track)
            self.__append(new_track)
            track._clear_dirty()  # pylint: disable=protected-access
            return new_track
        except Exception:
            # do not do self.remove. If we try to upload the same track to gpsies,
            # gpsies will assign the same trackid, and we come here. __tracks will
            # only hold the first uploaded track, and remove would remove that
            # instance instead of this one.
            # TODO: do we have a unittest for that case?
            self.__tracks = [x for x in self.__tracks if x is not new_track]
            with self._decouple():
                new_track.id_in_backend = None
                new_track._set_backend(None)  # pylint: disable=protected-access
            raise

    def _new_ident(self, track) ->str:
        """Create an id for track.

        Returns:
            The new ident. If the backend does not
            create an ident in advance, return None. Such
            backends will return a new ident after writing.

        """

    def _rewrite(self, track, changes):
        """Rewrite the full track.

        Used only by Track when things change.

        """
        self._current_track = track
        assert track.backend is self
        assert self._has_item(track.id_in_backend), '{}: its id_in_backend {} is not in {}'.format(
            track, track.id_in_backend, ' / '.join(str(x) for x in self))
        assert track._dirty  # pylint: disable=protected-access

        needs_full_save = self._needs_full_save(changes)

        self.matches(track, '_rewrite')
        if needs_full_save:
            new_id = self._write_all(track)
            track.id_in_backend = new_id
        else:
            for change in changes:
                _ = change.split(':')
                write_name = '_write_{}'.format(_[0])
                if len(_) == 1:
                    getattr(self, write_name)(track)
                elif len(_) == 2:
                    getattr(self, write_name)(track, _[1])
                else:
                    raise Exception('dirty {} got too many arguments:{}'.format(write_name, _[1:]))

    def _write_all(self, track) ->str:
        """the actual implementation for the concrete Backend.

        Writes the entire Track.

        Returns:
            The new id_in_backend

        """
        raise NotImplementedError()

    def remove(self, value) ->None:
        """
        Remove track. This can also be done for tracks not passing the current match function.

        Args:
            value: If it is not an :class:`~gpxity.Track`, :meth:`remove` looks
                it up by doing :literal:`self[value]`

        """
        track = value if hasattr(value, 'id_in_backend') else self[value]
        self._current_track = track
        if track.id_in_backend:
            self._remove_ident(track.id_in_backend)
        with self._decouple():
            track._set_backend(None)  # pylint: disable=protected-access
            try:
                self.__tracks = [x for x in self.__tracks if x.id_in_backend != track.id_in_backend]
            except ValueError:
                pass

    def _remove_ident(self, ident: str) ->None:
        """backend dependent implementation."""
        raise NotImplementedError()

    def _lifetrack_start(self, track, points):
        """Modelled after MapMyTracks. I hope this matches other services too.

        This will always produce a new track in the backend.supported

        If the backend does not support lifetrack, just add the points
        to the track.

        Args:
            track(Track): Holds initial data
            points: If None, stop tracking. Otherwise, start tracking
                and add points.

        For details see :meth:`Track.track() <gpxity.Track.track>`.

        """
        raise NotImplementedError()

    def _lifetrack_update(self, track, points):
        """If the backend does not support lifetrack, just add the points to the track.

        Args:
            track(Track): Holds initial data
            points: If None, stop tracking. Otherwise, start tracking
                and add points.

        For details see :meth:`Track.track() <gpxity.Track.track>`.

        """
        raise NotImplementedError()

    def _lifetrack_end(self, track):
        """If the backend does not support lifetrack, do nothing."""
        raise NotImplementedError()

    def remove_all(self):
        """Remove all tracks we know about.

        If their :attr:`id_in_backend` has meanwhile been changed
        through another backend instance or another process, we
        cannot find it anymore. We do **not** rescan all tracks in the backend.
        If you want to make sure it will be empty, call :meth:`scan` first.

        If you use a match function, only matching tracks will be removed."""
        for track in list(self):
            if self.matches(track):
                self.remove(track)

    def destroy(self):
        """If `cleanup` was set at init time, removes all tracks.

        Some backends (example: :class:`Directory <gpxity.Directory.destroy>`)

        may also remove the account (or directory). See also :meth:`remove_all`."""
        if self._cleanup:
            self.remove_all()

    def __contains__(self, value) ->bool:
        """value is either an a track or a track id.

        Does NOT load tracks, only checks what is already known.

        Returns:
            True if we have the item

        """
        self._scan()
        return self._has_item(value)

    def _has_item(self, index) ->bool:
        """like __contains__ but for internal use: does not call _scan first.

        Must not call self._scan.

        Returns:
            True if we have the item

        """
        if hasattr(index, 'id_in_backend') and index in self.__tracks:
            return True
        if isinstance(index, str) and index in [x.id_in_backend for x in self.__tracks]:
            return True
        return False

    def __getitem__(self, index):
        """Allow accesses like alist[a_id].

        Do not call this when implementing a backend because this always calls scan() first.

        Instead use :meth:`_has_item`.

        Returns:
            the track

        """
        self._scan()
        if isinstance(index, int):
            return self.__tracks[index]
        for _ in self.__tracks:
            if _ is index or _.id_in_backend == index:
                return _
        raise IndexError

    def __len__(self) ->int:
        """do not call this when implementing a backend because this calls scan().

        Returns:
            the length

        """
        self._scan()
        return len(self.__tracks)

    def real_len(self) ->int:
        """len(backend) without calling scan() first.

        Returns:

            the length"""
        return len(self.__tracks)

    def __append(self, track):
        """Append a track to the cached list."""
        self._current_track = track
        if track.id_in_backend is not None and not isinstance(track.id_in_backend, str):
            raise Exception('{}: id_in_backend must be str'.format(track))
        if track.id_in_backend is not None and any(x.id_in_backend == track.id_in_backend for x in self.__tracks):
            # cannot do "in self" because we are not decoupled, so that would call _scan()
            raise ValueError(
                'Backend.append(track): its id_in_backend {} is already in list: Track={}, list={}'.format(
                    track.id_in_backend, self[track.id_in_backend], self.__tracks))
        self.matches(track, 'append')
        self.__tracks.append(track)

    def __repr__(self):
        """do not call len(self) because that does things.

        Returns:
            The repr str

        """
        dirname = ''
        if self.auth:
            dirname = self.auth[0] or ''
        result = '{}({} in {}{})'.format(
            self.__class__.__name__, len(self.__tracks), self.url, dirname)
        return result

    def __enter__(self):
        """See class docstring.

        Returns:
            self

        """
        return self

    def __exit__(self, exc_type, exc_value, trback):
        """See class docstring."""
        self.destroy()

    def __iter__(self):
        """See class docstring.

        Returns:
            iterator over tracks

        """
        self._scan()
        return iter(self.__tracks)

    def __eq__(self, other) ->bool:
        """True if both backends have the same tracks.

        Returns:
            True if both backends have the same tracks

        """
        self._scan()
        other._scan()  # pylint: disable=protected-access
        return {x.key() for x in self} == {x.key() for x in other}

    def __copy(self, other_tracks, remove, dry_run):
        """Copy other_tracks into self. Used only by self.merge().

        Returns:
            verbose messages

        """
        result = list()
        for old_track in other_tracks:
            if not dry_run:
                new_track = self.add(old_track)
            result.append('{} {} -> {}'.format(
                'blind move' if remove else 'blind copy', old_track.identifier(),
                '' if dry_run else new_track.identifier()))
            if remove and not dry_run:
                old_track.remove()
        return result

    def merge(self, other, remove: bool = False, dry_run: bool = False, copy: bool = False) ->list:
        """merge other backend or a single track into this one.

        If two tracks have identical points, or-ify their other attributes.
        Args:
            other: The backend or a single track to be merged
            remove: If True, remove merged tracks
            dry_run: If True, do not really merge or remove
            copy: Do not try to find a matching track, just copy other into this Backend

        Returns:
            list(str) A list of messages for verbose output

        """
        # pylint: disable=too-many-branches,too-many-locals
        # TODO: test for dry_run
        # TODO: test for merging single track
        # TODO: test for merging a backend or a track with itself. Where
        # they may be identical instantiations or not. For all backends.
        result = list()
        other_tracks = collect_tracks(other)
        if copy:
            return self.__copy(other_tracks, remove, dry_run)

        src_dict = defaultdict(list)
        dst_dict = dict()
        for self_track in self:
            _ = self_track.points_hash()
            if _ not in dst_dict:
                dst_dict[_] = self_track
            else:
                src_dict[_].append(self_track)
        for _ in other_tracks:
            if _.backend.identifier() != self.identifier():
                src_dict[_.points_hash()].append(_)

        # 1. get all tracks existing only in other
        for point_hash in sorted(src_dict.keys() - dst_dict.keys()):
            for old_track in src_dict[point_hash]:
                if not dry_run:
                    new_track = self.add(old_track)
                result.append('{} {} -> {}'.format(
                    'move' if remove else 'copy', old_track.identifier(),
                    new_track.identifier() if new_track else self.identifier()))
                if remove and not dry_run:
                    old_track.remove()

        # 2. merge the rest
        for point_hash in src_dict.keys() & dst_dict.keys():
            target = dst_dict[point_hash]
            for source in src_dict[point_hash]:
                result.extend(target.merge(source, remove=remove, dry_run=dry_run))
        return result

    @staticmethod
    def _html_encode(value) ->str:
        """encode str to something gpies.com accepts.

        Returns:
            the encoded value

        """
        if value is None:
            return ''
        return value.encode('ascii', 'xmlcharrefreplace').decode()

    def flush(self):
        """Some backends delay actual writing. This enforces writing.

        Currently, only the Mailer backend can delay, it will bundle all
        mailed tracks into one mail instead of sending separate mails
        for every track. Needed for lifetracking.

        """

    @classmethod
    def is_disabled(cls) ->bool:
        """True if this backend is disabled by env variable GPXITY_DISABLE_BACKENDS.

        This variable is a comma separated list of Backend class names.

        Returns:
            True if disabled

        """
        disabled = os.getenv('GPXITY_DISABLE_BACKENDS')
        return disabled and cls.__name__ in disabled.split(',')



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.Authenticate`."""

import os
import tempfile
from configparser import ConfigParser

__all__ = ['Authenticate']


class Authenticate:

    """
    Get password and / or Url from auth.cfg.

    If nothing is useable, sets them to None.

    auth.cfg is expected in :literal:`~/.config/Gpxity/auth.cfg`

    .. DANGER::
       auth.cfg is not encrypted. Better not use this unless you know what you are doing!

    Args:
        cls (Backend): The class of the backend
        username (str): For the wanted account in the backend

    Attributes:
        path (str): The name for the auth file. Class variable, to be changed before
          Authenticate() is instantiated.

        auth (tuple(str,str)): (username, password). Both are either str or None.

        url: If given, overrides the url given to the backend

    For every specific account in a backend, auth.cfg has a section. Its name is
    case sensitive: The ClassName must match exactly.

      * [ClassName.username]

    A section can define
      * Password
      * Url

    An example for user gpxitytest on gpsies.com:

    .. code-block:: guess

        [GPSIES:gpxitytest]
        Password = the_unencrypted_password

    A mail account:

    .. code-block:: guess

        [Mailer:gpxitytest]
        Url = tester@test.test


    """

    # pylint: disable=too-few-public-methods

    path = '~/.config/Gpxity/auth.cfg'

    def __init__(self, cls, username: str = None):
        """See class docstring."""

        self.cls = cls
        self.__username = username
        self.auth = (None, None)
        self.section = dict()
        if 'Directory' in cls.__name__ and username.startswith('gpxitytest'):
            self.url = tempfile.mkdtemp(prefix='gpxity')
            return
        self.__path = os.path.expanduser(self.path)
        with open(self.__path) as auth_file:
            self._parse_config(auth_file.read())
        return

    def _parse_config(self, data):
        """try to use data."""

        password = url = None

        config = ConfigParser()
        config.read_string(data)

        config_key = '{}:{}'.format(self.cls.__name__, self.__username)
        try:
            section = config[config_key]
        except KeyError:
            raise KeyError('Section [{}] not found in {}'.format(config_key, self.__path))
        if 'Password' in section:
            password = section['Password']
        if 'Url' in section:
            url = section['Url']

        self.auth = (self.__username, password)
        self.url = url
        self.section = section



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This implements :class:`gpxity.ServerDirectory`."""


from .directory import Directory, Backup

__all__ = ['ServerDirectory']


class ServerDirectory(Directory):

    """Like :class:`Directory` but the track ids are different: Just a number.

    A new id is generated by adding 1 to the highest existing id.

    The symbolic links per YYYY/MM use the title of the track as link name.

    """

    # pylint: disable=abstract-method

    skip_test = True

    def _new_ident(self, _) ->str:
        """Build a unique id for track.

        Returns:
            A new unique id

        """
        try:
            return str(max(int(x) for x in self._list_gpx()) + 1)
        except ValueError:
            return '1'

    def _write_all(self, track) ->str:
        """save full gpx track. If id_in_backend is defined, keep it.

        Returns:
            The new id_in_backend

        """
        new_ident = track.id_in_backend or self._new_ident(track)

        with Backup(track):
            track.id_in_backend = new_ident
            with open(self.gpx_path(new_ident), 'w', encoding='utf-8') as out_file:
                out_file.write(track.to_xml())
            self._set_filetime(track)

        return new_ident



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""Defines :class:`gpxity.Directory`."""


import os
import datetime
import tempfile
from collections import defaultdict

import gpxpy.gpxfield as mod_gpxfield

from .. import Backend, Track
from ..util import remove_directory

__all__ = ['Directory', 'Backup']


class Backup:

    """A context manager making a backup of the gpx file.

    If an exception happened, restore the backup.
    Otherwise, remove it again.

    """

    # pylint: disable=too-few-public-methods

    def __init__(self, track):
        """See class docstring."""
        self.track = track
        self.old_id = track.id_in_backend
        self.old_pathname = None
        if self.old_id is not None:
            self.old_pathname = track.backend.gpx_path(self.old_id)
            if os.path.exists(self.old_pathname):
                os.rename(self.old_pathname, self.old_pathname + '.old')

    def __enter__(self):
        """See class docstring.

        Returns:
            self

        """
        return self

    def __exit__(self, exc_type, exc_value, trback):
        """See class docstring."""
        if exc_value:
            self.undo_rename()
            with self.track._decouple():  # pylint: disable=protected-access
                self.track.id_in_backend = self.old_id
        else:
            if self.old_pathname is not None:
                if os.path.exists(self.old_pathname + '.old'):
                    os.remove(self.old_pathname + '.old')

    def undo_rename(self):
        """if something failed, undo change of file name and restore old file."""
        if self.old_pathname is not None:
            if os.path.exists(self.old_pathname):
                os.remove(self.old_pathname)
            if os.path.exists(self.old_pathname + '.old'):
                os.rename(self.old_pathname + '.old', self.old_pathname)


class Directory(Backend):

    """Uses a directory for storage. The filename minus the .gpx ending is used as the track id.

    If the track has a title but no storage id yet, use the title as storage id.
    Make the storage id unique by attaching a number if needed.
    An track without title gets a random name.

    The main directory (given by :attr:`Directory.url <gpxity.backend.Backend.url>`) will have
    subdirectories YYYY/MM (year/month) with only the tracks for one month.
    Those are symbolic links to the main file and have the same file name.

    If :meth:`~gpxity.backend.Backend.save` is given a value for ident, this
    is used as id, the file name will be :literal:`id.gpx`.
    Otherwise, this backend uses :attr:`Track.title <gpxity.Track.title>` for the id.
    If a track has no title, it uses a random sequence of characters.
    Changing the title also changes the id.

    Args:
        url (str): a directory. If no Url is given, either here or through auth, use a unique
            temporary directory named
            :attr:`prefix`.X where X are some random characters.
            If the directory does not exist, it is created.
        auth (str): You can use this as in every backend to define Url= in auth.cfg
        cleanup (bool): If True, :meth:`destroy` will remove all tracks. If url was
            not given, it will also remove the directory.
        prefix: The prefix for a temporary directory path. Must not be given if url is given.

    Attributes:
        prefix (str):  Class attribute, may be changed. The default prefix for
            temporary directories. Default value is :literal:`gpxity.`
        fs_encoding (str): The encoding for file system names. By default, we
            expect the file system being able to handle arbitrary UTF-8 encoded names
            except character '/' and special names '.' and '..'. If needed, we will introduce
            new possible values for fs_encoding like perhaps 'windows'. Gpxity will **never**
            support any other character set but UTF-8.
            Note that :attr:`fs_encoding` is independent of the platform we are running on - we
            might use a network file system.
        is_temporary (bool): True if no Url was given and we created a temporary directory

    """

    # skip_test = True
    # pylint: disable=abstract-method

    prefix = 'gpxity.'

    def __init__(self, url=None, auth=None, cleanup=False, prefix: str = None):
        """See class docstring."""
        self.fs_encoding = None
        if prefix is None:
            prefix = self.__class__.prefix
        elif url:
            raise Exception('Directory does not accept both url and prefix')
        super(Directory, self).__init__(url=url, auth=auth, cleanup=cleanup)
        self.is_temporary = not bool(self.url)
        if self.is_temporary:
            self.url = tempfile.mkdtemp(prefix=prefix)
        if not os.path.exists(self.url):
            os.makedirs(self.url)
        self._symlinks = defaultdict(list)
        self._load_symlinks()

    def identifier(self, track=None) ->str:
        """Used for formatting strings. Must be unique.

        Returns:

            a unique identifier"""
        result = self.url
        if result:
            if result.startswith('./') or result == '.':
                result = result[2:]
        if track:
            if result:
                result += '/' + track.id_in_backend
            else:
                result = track.id_in_backend
        return result

    @property
    def legal_categories(self):
        """
        A list with all legal categories.

        Returns: list(str)
            all legal values for category for this backend.

        """
        return Track.legal_categories

    def decode_category(self, value: str) ->str:
        """Not needed for directory, this is always the internal value.

        Returns:
            the decoded category

        """
        return value

    def encode_category(self, value: str) ->str:
        """Not needed for directory, this is always the internal value.

        Returns:
            the encoded category

        """
        return value

    def _load_symlinks(self, directory=None):
        """scan the subdirectories with the symlinks.

        If the content of a track changes, the symlinks might have to
        be adapted. But we do not know the name of the existing symlink anymore.

        So just scan them all and assign them to id_in_backend."""
        if directory is None:
            directory = self.url
        for dirpath, _, filenames in os.walk(directory):
            for filename in filenames:
                full_name = os.path.join(dirpath, filename)
                if os.path.islink(full_name):
                    if os.path.exists(full_name):
                        target = os.readlink(full_name)
                        gpx_target = os.path.basename(target)
                        if gpx_target.endswith('.gpx'):
                            # it really should ...
                            gpx_target = gpx_target[:-4]
                        if full_name not in self._symlinks[gpx_target]:
                            self._symlinks[gpx_target].append(full_name)
                    else:
                        os.remove(full_name)

    def _new_id_from(self, ident_proposal: str) ->str:
        """Return not yet existant file name.

        Args:
            ident_proposal: If this proposal does not lead to a valid ident, create unique random ident.

        Returns:
            The new unique ident

        """
        value = self._sanitize_name(ident_proposal)
        if not value:
            value = os.path.basename(tempfile.NamedTemporaryFile(dir=self.url, prefix='').name)
        return self._make_ident_unique(value)

    @staticmethod
    def _make_path_unique(value) ->str:
        """If the file name already exists, apply a serial number.

        If value ends with .gpx, put the serial number in front of that.

        Returns:
            the unique path name

        """
        ctr = 0
        unique_value = value
        while os.path.exists(unique_value):
            ctr += 1
            if value.endswith('.gpx'):
                unique_value = '{}.{}.gpx'.format(value[:-4], ctr)
            else:
                unique_value = '{}.{}'.format(value, ctr)
        return unique_value

    def _make_ident_unique(self, value):
        """Return a unique ident."""
        path = Directory._make_path_unique(os.path.join(self.url, value + '.gpx'))
        return os.path.basename(path)[:-4]

    def _sanitize_name(self, value) ->str:
        """Change it to legal file name characters.

        Returns:
            the sanitized name

        """
        if value is None:
            return None
        if self.fs_encoding is not None:
            raise Exception('No support for fs_encoding={}'.format(self.fs_encoding))
        return value.replace('/', '_')

    def destroy(self):
        """If `cleanup` was set at init time, removes all tracks.
        If :attr:`~gpxity.Directory.url` was set at init time,
        also removes the directory."""
        super(Directory, self).destroy()
        if self._cleanup:
            self.remove_all()
            if self.is_temporary:
                remove_directory(self.url)

    def gpx_path(self, ident) ->str:
        """The full path name for the local copy of a track.

        Returns:
            The full path name

        """
        assert isinstance(ident, str), '{} must be str'.format(ident)
        return os.path.join(self.url, '{}.gpx'.format(ident))

    def _list_gpx(self):
        """return a generator of all gpx files, with .gpx removed.

        Returns:
            A list of all gpx file names with .gpx removed

        """
        gpx_names = (x for x in os.listdir(self.url) if x.endswith('.gpx'))
        return (x.replace('.gpx', '') for x in gpx_names)

    @staticmethod
    def _get_field(data, name) ->str:
        """Get xml field out of data.

        Returns:
            The xml field

        """
        start_html = '<{}>'.format(name)
        end_html = '</{}>'.format(name)
        data = data.split(end_html)
        if len(data) > 1:
            data = data[0]
            data = data.split(start_html)
            if len(data) > 1:
                data = data[-1]
                if start_html not in data:
                    return data
        return None

    def _enrich_with_headers(self, track):
        """Quick scan of file for getting some header fields."""
        with open(self.gpx_path(track.id_in_backend)) as raw_file:
            data = raw_file.read(10000)
            parts = data.split('<trk>')
            if len(parts) > 1:
                # pylint: disable=protected-access
                raw_data = parts[0].split('extensions')[0]
                _ = self._get_field(raw_data, 'name')
                if _ is not None:
                    track._header_data['title'] = _
                _ = self._get_field(raw_data, 'desc')
                if _ is not None:
                    track._header_data['description'] = _
                rest_kw = list()
                _ = self._get_field(raw_data, 'keywords')
                if _:
                    for keyword in (x.strip() for x in _.split(',')):
                        if keyword == 'Status:public':
                            track._header_data['public'] = True
                        elif keyword == 'Status:private':
                            track._header_data['public'] = False
                        elif keyword.startswith('Category:'):
                            track._header_data['category'] = keyword.split(':')[1].strip()
                        else:
                            rest_kw.append(keyword)
                    track._header_data['keywords'] = sorted(rest_kw)

                _ = self._get_field(parts[1], 'time')
                if _ is not None:
                    track._header_data['time'] = mod_gpxfield.parse_time(_)

    def _yield_tracks(self):
        """get all tracks for this user."""
        self._symlinks = defaultdict(list)
        self._load_symlinks()
        for _ in self._list_gpx():
            track = self._found_track(_)
            self._enrich_with_headers(track)
            yield track

    def get_time(self) ->datetime.datetime:
        """get server time as a Linux timestamp.

        Returns:
            The server time

        """
        return datetime.datetime.now()

    def _read_all(self, track):
        """fill the track with all its data from source."""
        assert track.id_in_backend
        with open(self.gpx_path(track.id_in_backend), encoding='utf-8') as in_file:
            track.parse(in_file)

    def _remove_symlinks(self, ident: str):
        """Remove its symlinks, empty symlink parent directories."""
        for symlink in self._symlinks[ident]:
            if os.path.exists(symlink):
                os.remove(symlink)
            symlink_dir = os.path.split(symlink)[0]
            try:
                os.removedirs(symlink_dir)
            except OSError:
                pass
        self._symlinks[ident] = list()

    def _remove_ident(self, ident: str):
        """Remove its symlinks and the file, in this order."""
        self._remove_symlinks(ident)
        gpx_file = self.gpx_path(ident)
        if os.path.exists(gpx_file):
            os.remove(gpx_file)

    def _symlink_path(self, track) ->str:
        """The path for the speaking symbolic link: YYYY/MM/title.gpx.

        Missing directories YYYY/MM are created.

        Returns:
            The path

        """
        ident = track.id_in_backend
        time = datetime.datetime.fromtimestamp(os.path.getmtime(self.gpx_path(ident)))
        by_month_dir = os.path.join(self.url, '{}'.format(time.year), '{:02}'.format(time.month))
        if not os.path.exists(by_month_dir):
            os.makedirs(by_month_dir)
        else:
            # make sure there is no dead symlink with our wanted name.
            self._load_symlinks(by_month_dir)
        name = track.title or ident
        return self._make_path_unique(os.path.join(by_month_dir, self._sanitize_name(name)))

    def _new_ident(self, track) ->str:
        """Create an id for track.

        Returns: The new ident.

        """
        ident = track.id_in_backend
        if ident is None:
            ident = self._new_id_from(None)
        return ident

    def _make_symlinks(self, track):
        """Make all symlinks for track."""
        ident = track.id_in_backend
        gpx_pathname = self.gpx_path(ident)
        link_name = self._symlink_path(track)
        basename = os.path.basename(gpx_pathname)
        link_target = os.path.join('..', '..', basename)
        os.symlink(link_target, link_name)
        if link_name not in self._symlinks[ident]:
            self._symlinks[ident].append(link_name)

    def _set_filetime(self, track):
        """Set the file modification time to track start time.
        If the track has no start time, do nothing."""
        time = track.time
        if time:
            _ = self.gpx_path(track.id_in_backend)
            os.utime(_, (time.timestamp(), time.timestamp()))

    def _change_id(self, track, new_ident: str):
        """Change the id in the backend."""
        assert track.id_in_backend != new_ident
        unique_id = self._new_id_from(new_ident)
        self._remove_symlinks(track.id_in_backend)
        os.rename(self.gpx_path(track.id_in_backend), self.gpx_path(unique_id))
        track.id_in_backend = unique_id
        assert any(x is track for x in self._Backend__tracks)
        self._make_symlinks(track)

    def _write_all(self, track) ->str:
        """save full gpx track.

        Since the file name uses title and title may have changed,
        compute new file name and remove the old files. We also adapt track.id_in_backend.

        Returns:
            the new track.id_in_backend

        """
        old_ident = track.id_in_backend
        new_ident = self._new_ident(track)

        with Backup(track):
            track.id_in_backend = new_ident
            with open(self.gpx_path(new_ident), 'w', encoding='utf-8') as out_file:
                out_file.write(track.to_xml())
            self._set_filetime(track)

        if old_ident and new_ident != old_ident:
            self._remove_symlinks(old_ident)
            self._make_symlinks(track)
        return new_ident



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
This implements only a minimum of what MMT can do.

Upload entire tracks and extend a track. That is
what oruxmaps does - see examples/mmt_server.

TrackMMT is used to test mmt_server.

"""

from .mmt import MMT
from ..track import Track

__all__ = ['TrackMMT']


class TrackMMT(MMT):

    """This is a minimal implementation, it only supports listing and retrieving tracks and life tracking.

    This is used for testing bin/gpxity_server which in turn is used to
    receive life tracking data from smartphone apps like
    oruxmaps.

    """

    # pylint: disable=abstract-method

    legal_categories = Track.legal_categories

    def _yield_tracks(self):
        """not implemented."""
        raise NotImplementedError()

    def _write_attribute(self, track, attribute):
        """not implemented."""
        raise NotImplementedError()

    def _write_title(self, track):
        """not implemented."""
        raise NotImplementedError()

    def _write_description(self, track):
        """not implemented."""
        raise NotImplementedError()

    def _write_public(self, track):
        """not implemented."""
        raise NotImplementedError()

    def _write_category(self, track):
        """not implemented."""
        raise NotImplementedError()

    def _write_add_keywords(self, track, values):
        """not implemented."""
        raise NotImplementedError()

    def _write_remove_keywords(self, track, values):
        """not implemented."""
        raise NotImplementedError()

    def _remove_ident(self, ident: str) ->None:
        """backend dependent implementation."""
        raise NotImplementedError()

    @property
    def is_free_account(self):
        """Our own local server can do lifeftracking.

        Returns:
            False

        """
        return False



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""implements :class:`gpxity.backends.test.test_backends.TestBackends` for all backends."""

import os
import pwd
import time
import datetime
import random
import tempfile

from unittest import skip

from .basic import BasicTest
from .. import Directory, MMT, GPSIES, ServerDirectory, TrackMMT, Mailer
from ...auth import Authenticate
from ... import Track, Lifetrack

# pylint: disable=attribute-defined-outside-init


class TestBackends(BasicTest):

    """Are the :literal:`supported_` attributes set correctly?."""

    def test_supported(self):
        """Check values in supported for all backends."""
        expect_unsupported = dict()
        expect_unsupported[Directory] = {
            'lifetrack', 'lifetrack_end', 'write_add_keywords', 'write_remove_keywords', 'write_gpx',
            'write_category', 'write_description', 'write_public', 'write_title'}
        expect_unsupported[ServerDirectory] = {'lifetrack'}
        expect_unsupported[MMT] = {'write_gpx'}
        expect_unsupported[GPSIES] = {
            'get_time', 'lifetrack', 'lifetrack_end', 'write_add_keywords', 'write_remove_keywords', 'write_gpx'}
        expect_unsupported[TrackMMT] = {
            'scan', 'remove', 'write_gpx',
            'write_title', 'write_description', 'write_public',
            'write_category', 'write_add_keywords',
            'write_remove_keywords'}
        expect_unsupported[Mailer] = {
            'scan', 'remove', 'get_time',
            'write_title', 'write_description', 'write_public',
            'write_category', 'write_add_keywords', 'write_gpx',
            'write_remove_keywords', 'lifetrack'}
        for cls in self._find_backend_classes():
            self.logger.debug('%s %s', cls, cls.supported)
            self.logger.debug('%s %s', cls, expect_unsupported[cls])
            with self.subTest(' {}'.format(cls.__name__)):
                self.assertTrue(cls.supported & expect_unsupported[cls] == set())
                self.assertEqual(sorted(cls.supported | expect_unsupported[cls]), sorted(cls.full_support))

    def test_save_empty(self):
        """Save empty track."""
        for cls in self._find_backend_classes():
            if 'write' not in cls.supported:
                continue
            with self.subTest(' {}'.format(cls.__name__)):
                can_remove = 'remove' in cls.supported
                with self.temp_backend(cls, cleanup=can_remove, clear_first=can_remove) as backend:
                    track = Track()
                    if cls in (MMT, TrackMMT, GPSIES):
                        with self.assertRaises(cls.BackendException):
                            backend.add(track)
                    else:
                        self.assertIsNotNone(backend.add(track))

    def test_directory_backend(self):
        """Manipulate backend."""
        track = self.create_test_track()
        with Directory(cleanup=True) as directory1:
            with Directory(cleanup=True) as directory2:
                saved = directory1.add(track)
                self.assertEqual(len(directory1), 1)
                self.assertEqual(saved.backend, directory1)
                directory1.add(track.clone())
                self.assertEqual(len(directory1), 2)
                directory2.add(track)
                self.assertEqual(len(directory2), 1)
                directory2.scan()
                self.assertEqual(len(directory2), 1)

    def test_slow_duplicate_tracks(self):
        """What happens if we save the same track twice?."""
        for cls in self._find_backend_classes():
            if 'remove' in cls.supported and 'write' in cls.supported:
                with self.subTest(' {}'.format(cls.__name__)):
                    with self.temp_backend(cls) as backend:
                        track = self.create_test_track()
                        backend.add(track)
                        self.assertEqual(len(backend), 1)
                        with self.assertRaises(ValueError):
                            backend.add(track)
                        self.assertEqual(len(backend), 1)
                        if cls is GPSIES:
                            # if the same track data is uploaded again, we get the same id_in_backend.
                            with self.assertRaises(ValueError):
                                backend.add(track.clone())
                            self.assertEqual(len(backend), 1)
                        else:
                            backend.add(track.clone())
                            self.assertEqual(len(backend), 2)

    def test_open_wrong_username(self):
        """Open backends with username missing in auth.cfg."""
        for cls in self._find_backend_classes():
            with self.subTest(' {}'.format(cls.__name__)):
                with self.assertRaises(KeyError):
                    self.setup_backend(cls, username='wrong')

    def test_open_wrong_password(self):
        """Open backends with wrong password."""
        for cls in self._find_backend_classes():
            if 'scan' not in cls.supported:
                continue
            with self.subTest(' {}'.format(cls.__name__)):
                if not issubclass(cls, Directory):
                    with self.assertRaises(cls.BackendException):
                        self.setup_backend(cls, username='wrong_password')

    def test_match(self):
        """test backend match function.

        Returns:
            None

        """
        def match_date(track) ->str:
            """match against a date.

            Returns:
                None if match else an error message
            """
            if track.time < datetime.datetime(year=2016, month=9, day=5):
                return 'time {} is before {}'.format(track.time, '2016-09-05')
            return None
        cls = Directory
        with self.subTest(' {}'.format(cls.__name__)):
            with self.temp_backend(cls, count=3) as backend:
                for idx, _ in enumerate(backend):
                    _.adjust_time(datetime.timedelta(hours=idx))
                new_track = backend[0].clone()
                self.assertIsNotNone(match_date(new_track))
                self.assertEqual(len(backend), 3)
                backend.match = match_date
                self.assertEqual(len(backend), 1)
                with self.assertRaises(cls.NoMatch):
                    backend.add(new_track)
                self.assertEqual(len(backend), 1)
                orig_time = backend[0].time
                delta = datetime.timedelta(days=-5)
                with self.assertRaises(cls.NoMatch):
                    backend[0].adjust_time(delta)
                self.assertEqual(len(backend), 1)
                self.assertEqual(orig_time + delta, backend[0].time)

    def test_z9_create_backend(self):
        """Test creation of a backend."""
        for cls in self._find_backend_classes():
            if 'remove' in cls.supported and 'get_time' in cls.supported:
                with self.subTest(' {}'.format(cls.__name__)):
                    with self.temp_backend(cls, count=3) as backend:
                        self.assertEqual(len(backend), 3)
                        first_time = backend.get_time()
                        time.sleep(2)
                        second_time = backend.get_time()
                        total_seconds = (second_time - first_time).total_seconds()
                        self.assertTrue(1 < total_seconds < 8, 'Time difference should be {}, is {}-{}={}'.format(
                            2, second_time, first_time, second_time - first_time))

    def test_slow_write_remoteattr(self):
        """If we change title, description, public, category in track, is the backend updated?."""
        for cls in self._find_backend_classes():
            if 'remove' in cls.supported:
                with self.subTest(' {}'.format(cls.__name__)):
                    with self.temp_backend(cls, count=1, category='Horse riding') as backend:
                        track = backend[0]
                        first_public = track.public
                        first_title = track.title
                        first_description = track.description
                        first_category = track.category
                        self.assertEqual(first_category, 'Horse riding')
                        self.assertFalse(track.public)
                        track.public = True
                        self.assertTrue(track.public)
                        track.title = 'A new title'
                        self.assertEqual(track.title, 'A new title')
                        track.description = 'A new description'
                        track.category = 'Cycling'
                        # make sure there is no cache in the way
                        backend2 = self.clone_backend(backend)
                        track2 = backend2[0]
                        self.assertEqualTracks(track, track2, with_category=False)
                        self.assertNotEqual(first_public, track2.public)
                        self.assertNotEqual(first_title, track2.title)
                        self.assertNotEqual(first_description, track2.description)
                        self.assertNotEqual(first_category, track2.category)

    def xtest_gpsies_bug(self):
        """We have this bug only sometimes: title, category or time will be wrong in track2.
        Workaround is in GPSIES._edit."""
        for _ in range(20):
            with self.temp_backend(GPSIES, count=1, category='Horse riding') as backend:
                track = backend[0]
                track.title = 'A new title'
                track.description = 'A new description'
                track.category = 'Cycling'
                # make sure there is no cache in the way
                backend2 = self.clone_backend(backend)
                track2 = backend2[0]
                self.assertEqualTracks(track, track2, with_category=True)

    @skip
    def test_zz_all_category(self):
        """can we up- and download all values for :attr:`Track.category`?."""
        category_count = len(Track.legal_categories)
        backends = [
            self.setup_backend(x, count=category_count, clear_first=True)
            for x in self._find_backend_classes() if 'remove' in x.supported]
        copies = [self.clone_backend(x) for x in backends]
        try:
            first_backend = copies[0]
            for other in copies[1:]:
                self.assertSameTracks(first_backend, other)
        finally:
            for backend in copies:
                backend.destroy()
            for backend in backends:
                backend.destroy()

    def test_z2_keywords(self):
        """save and load keywords.

        For now, all test keywords start with uppercase, avoiding MMT problems"""
        kw_a = 'A'
        kw_b = 'Berlin'
        kw_c = 'CamelCase'
        kw_d = 'D'  # self.unicode_string2

        for cls in self._find_backend_classes():
            if 'write' not in cls.supported or 'scan' not in cls.supported:
                continue
            with self.subTest(' {}'.format(cls.__name__)):
                is_mmt = cls.__name__ == 'MMT'
                with self.temp_backend(cls, clear_first=not is_mmt, cleanup=not is_mmt) as backend:
                    if not backend:
                        continue
                    track = backend[0]
                    track.keywords = list()
                    self.assertEqual(track.keywords, list())
                    track.keywords = ([kw_a, kw_b, kw_c])
                    track.remove_keywords(kw_b)
                    self.assertEqual(track.keywords, ([kw_a, kw_c]))
                    with self.assertRaises(Exception):
                        track.add_keywords('Category:whatever')
                    track.add_keywords(kw_d)
                    self.assertEqual(set(track.keywords), {kw_a, kw_c, kw_d})
                    backend2 = self.clone_backend(backend)
                    track2 = backend2[track.id_in_backend]
                    track2.remove_keywords(kw_d)
                    self.assertEqual(track2.keywords, ([kw_a, kw_c]))
                    self.assertEqual(track.keywords, ([kw_a, kw_c, kw_d]))
                    backend.scan()
                    self.assertEqual(track.keywords, ([kw_a, kw_c, kw_d]))
                    self.assertEqual(backend[track.id_in_backend].keywords, ([kw_a, kw_c]))
                    track.remove_keywords(kw_a)
                    # this is tricky: The current implementation assumes that track.keywords is
                    # current - which it is not. track still thinks kw_d is there but it has been
                    # removed by somebody else. MMT has a work-around for removing tracks which
                    # removes them all and re-adds all wanted. So we get kw_d back.
                    self.assertEqual(track.keywords, ([kw_c, kw_d]))
                    # track2.remove_keywords(kw_a)
                    track.remove_keywords(kw_c)
                    track.remove_keywords(kw_d)
                    backend.scan()
                    self.assertEqual(backend[0].keywords, list())

    def test_z_unicode(self):
        """Can we up- and download unicode characters in all text attributes?."""
        tstdescr = 'DESCRIPTION with ' + self.unicode_string1 + ' and ' + self.unicode_string2
        for cls in self._find_backend_classes():
            if 'remove' in cls.supported:
                with self.subTest(' {}'.format(cls.__name__)):
                    with self.temp_backend(cls, count=1) as backend:
                        backend2 = self.clone_backend(backend)
                        track = backend[0]
                        self.assertIsNotNone(track.backend)
                        track.title = 'Title ' + self.unicode_string1
                        self.assertIsNotNone(track.backend)
                        self.assertEqual(track.backend, backend)
                        backend2.scan()  # because backend2 does not know about changes thru backend
                        track2 = backend2[0]
                        # track and track2 may not be identical. If the original track
                        # contains gpx xml data ignored by MMT, it will not be in track2.
                        self.assertEqual(track.title, track2.title)
                        track.description = tstdescr
                        self.assertEqual(track.description, tstdescr)
                        if cls is Directory:
                            self.assertTrackFileContains(track, tstdescr)
                        backend2.scan()
                        self.assertEqual(backend2[0].description, tstdescr)
                        backend2.destroy()

    def test_change_points(self):
        """Can we change the points of a track?.

        For MMT this means re-uploading and removing the previous instance, so this

        is not always as trivial as it should be."""

    def test_slow_download_many(self):
        """Download many tracks."""
        many = 150
        backend = self.setup_backend(MMT, username='gpxstoragemany', count=many, cleanup=False, clear_first=True)
        self.assertEqual(len(backend), many)

    def test_duplicate_title(self):
        """two tracks having the same title."""
        for cls in self._find_backend_classes():
            if 'remove' in cls.supported:
                with self.subTest(' {}'.format(cls.__name__)):
                    with self.temp_backend(cls, count=2) as backend:
                        backend[0].title = 'TITLE'
                        backend[1].title = 'TITLE'

    def test_private(self):
        """Up- and download private tracks."""
        with self.temp_backend(Directory, count=5, category='Cycling') as local:
            track = Track(gpx=self._get_gpx_from_test_file('test2'))
            self.assertTrue(track.public)  # as defined in test2.gpx keywords
            track.public = False
            self.assertFalse(track.public)
            local.add(track)
            for cls in self._find_backend_classes():
                if 'remove' in cls.supported:
                    with self.subTest(' {}'.format(cls.__name__)):
                        with self.temp_backend(cls) as backend:
                            backend.merge(local)
                            for _ in backend:
                                self.assertFalse(_.public)
                            backend2 = self.clone_backend(backend)
                            with Directory(cleanup=True) as copy:
                                copy.merge(backend2)
                                self.assertSameTracks(local, copy)

    def test_merge(self):
        """merge backends."""
        with self.temp_backend(Directory, count=5) as source:

            with self.temp_backend(Directory, username='gpxitytest2', count=4) as sink:
                for _ in list(sink)[1:]:
                    _.adjust_time(datetime.timedelta(hours=100))
                sink.merge(source)
                self.assertEqual(len(source), 5)
                self.assertEqual(len(sink), 8)
                sink.merge(source, remove=True)
                self.assertEqual(len(source), 0)

    def test_scan(self):
        """some tests about Backend.scan()."""
        with self.temp_backend(Directory, count=5) as source:
            backend2 = self.clone_backend(source)
            track = self.create_test_track()
            backend2.add(track)
            self.assertEqual(len(backend2), 6)
            source.scan()  # because it cannot know backend2 added something

    def test_lifetrack_mmt(self):
        """test life tracking against a free account on mapmytracks.com."""
        with MMT(auth='gpxitytest') as uplink:
            self.assertTrue(uplink.is_free_account)
            life = Lifetrack(uplink)
            with self.assertRaises(Exception) as context:
                life.update(self._random_points())
            self.assertEqual(str(context.exception), 'Your free MMT account does not allow lifetracking')

    def test_lifetrack_local(self):
        """test life tracking against a local server."""
        if not TrackMMT.is_disabled() and not Mailer.is_disabled():
            with self.temp_backend(Directory) as serverdirectory:
                with self.lifetrackserver(servername='localhost', port=12398, directory=serverdirectory.url):
                    with TrackMMT(auth='gpxitytest') as uplink:
                        with Mailer(url=pwd.getpwuid(os.geteuid()).pw_name) as mailer:
                            mailer.min_interval = 5
                            life = Lifetrack([uplink, mailer])
                            points = self._random_points(100)
                            life.update(points[:50])
                            time.sleep(7)
                            life.update(points[50:])
                            life.end()
                            for target in life.targets:
                                track = target.track
                                new_id = track.id_in_backend
                                if 'scan' in target.backend.supported:
                                    self.assertIn(new_id, uplink)
                                    self.assertSameTracks(uplink, serverdirectory)
                            with self.assertRaises(NotImplementedError):
                                new_id in uplink  # pylint: disable=pointless-statement
                            self.assertEqual(len(mailer.history), 3)
                            self.assertIn('Lifetracking starts', mailer.history[0])
                            self.assertIn('Lifetracking continues', mailer.history[1])
                            self.assertIn('Lifetracking ends', mailer.history[2])

    def test_backend_dirty(self):
        """Track._dirty."""
        # pylint: disable=protected-access
        for cls in self._find_backend_classes():
            if 'scan' not in cls.supported or 'write' not in cls.supported:
                continue
            with self.temp_backend(cls, count=1) as backend:
                track = backend[0]
                with self.assertRaises(Exception):
                    track._dirty = False
                self.assertFalse(track._dirty)
                # version 1.1 should perhaps be a test on its own, see Track.to_xml()
                track.category = 'Driving'
                track._dirty = 'gpx'
                self.assertFalse(track._dirty)
                backend2 = self.clone_backend(backend)
                self.assertEqual(backend2[0].category, 'Driving')
                b2track = backend2[0]
                self.assertEqual(b2track.category, 'Driving')
                b2track.title = 'another new title'
                self.assertEqual(b2track.category, 'Driving')
                self.assertEqual(backend2[0].category, 'Driving')
                track.title = 'new title'
                self.assertEqual(track.category, 'Driving')
                backend3 = self.clone_backend(backend)
                self.assertEqual(backend3[0].category, 'Driving')
                self.assertFalse(track._dirty)
                with track.batch_changes():
                    track.title = 'new 2'
                    self.assertEqual(track._dirty, ['title'])
                self.assertFalse(track._dirty)
                with track.batch_changes():
                    track.title = 'new 3'
                    track.keywords = ['Something', 'something xlse']
                backend4 = self.clone_backend(backend)
                self.assertEqual(backend4[0].title, 'new 3')

    def test_directory_dirty(self):
        """test gpx._dirty where id_in_backend is not the default.

        Currently track._dirty = 'gpx' changes the file name which is wrong."""
        pass

    def test_directory(self):
        """directory creation/deletion."""
        with self.assertRaises(Exception):
            with Directory('url', prefix='x', cleanup=True):
                pass

        dir_a = Directory(cleanup=True)
        self.assertTrue(dir_a.is_temporary)
        a_url = dir_a.url
        self.assertTrue(os.path.exists(a_url))
        dir_a.destroy()
        self.assertFalse(os.path.exists(a_url))

        test_url = tempfile.mkdtemp() + '/'
        dir_b = Directory(url=test_url, cleanup=True)
        self.assertFalse(dir_b.is_temporary)
        self.assertTrue(dir_b.url == test_url)
        dir_b.destroy()
        self.assertTrue(os.path.exists(test_url))
        os.rmdir(test_url)

        dir_c = Directory(auth='gpxitytest2')
        auth_dir = Authenticate(Directory, 'gpxitytest2').url
        if not auth_dir.endswith('/'):
            auth_dir += '/'
        self.assertFalse(dir_c.is_temporary)
        self.assertNotEqual(dir_c.url, auth_dir)
        dir_c.destroy()
        self.assertTrue(os.path.exists(auth_dir))
        os.rmdir(auth_dir)

    def test_mmt_empty(self):
        """MMT refuses upload without a specific error message if there is no track point."""
        track = self.create_test_track()
        del track.gpx.tracks[0]
        with MMT(auth='gpxitytest', cleanup=True) as mmt:
            with self.assertRaises(mmt.BackendException):
                mmt.add(track)

    def test_setters(self):
        """For all Track attributes with setters, test if we can change them without changing something else."""
        for cls in self._find_backend_classes():
            if 'write' not in cls.supported or 'scan' not in cls.supported:
                continue
            with self.temp_backend(cls, count=1) as backend:
                track = backend[0]
                backend2 = self.clone_backend(backend)
                self.assertEqualTracks(track, backend2[0], with_category=False)
                test_values = {
                    'title': ('default title', 'Täst Titel'),
                    'description': ('default description', 'Täst description'),
                    'category': ('Driving', 'Rowing'), 'public': (True, False)}
                if cls is not GPSIES:
                    test_values['keywords'] = (['A', 'Hello Dolly', 'Whatever'], ['Something Else', 'Two'])
                for main in test_values:
                    for key, (default_value, _) in test_values.items():
                        if key != main:
                            setattr(track, key, default_value)
                    setattr(track, main, test_values[main][1])
                    backend2.scan()
                    for key, (default_value, _) in test_values.items():
                        if key != main:
                            self.assertEqual(getattr(backend2[0], key), default_value)
                    self.assertEqual(getattr(backend2[0], main), test_values[main][1])

    def test_keywords(self) ->None:
        """Test arbitrary keyword changes.

        Returns:
            None

        """

        def current():
            """The current keywords.

            Returns:
                The current keywords
            """
            if hasattr(backend, '_get_current_keywords'):
                return backend._get_current_keywords(track)  # pylint: disable=protected-access
            return track.keywords

        for cls in self._find_backend_classes():
            if 'write_add_keywords' not in cls.supported:
                continue
            with self.temp_backend(cls, count=1) as backend:
                backend2 = self.clone_backend(backend)
                track = backend[0]
                keywords = {
                    backend._encode_keyword(x)  # pylint: disable=protected-access
                    for x in self._random_keywords(count=50)}
                for _ in range(20):
                    self.assertEqual(current(), track.keywords)
                    add_keywords = set(random.sample(keywords, random.randint(0, 30)))
                    remove_keywords = set(random.sample(keywords, random.randint(0, 30)))
                    if not add_keywords & remove_keywords:
                        continue
                    expected_keywords = (set(track.keywords) | add_keywords) - remove_keywords
                    track.add_keywords(add_keywords)
                    self.assertEqual(current(), sorted(list(set(track.keywords) | add_keywords)))
                    track.remove_keywords(remove_keywords)
                    self.assertEqual(current(), sorted(expected_keywords))
                    self.assertEqual(sorted(expected_keywords), sorted(track.keywords))
                    backend2.scan()
                    self.assertEqual(sorted(expected_keywords), backend2[0].keywords)
                with track.batch_changes():
                    for _ in range(50):
                        add_keywords = set(random.sample(keywords, random.randint(0, 30)))
                        remove_keywords = set(random.sample(keywords, random.randint(0, 30)))
                        if not add_keywords & remove_keywords:
                            continue
                        expected_keywords = (set(track.keywords) | add_keywords) - remove_keywords
                        track.add_keywords(add_keywords)
                        track.remove_keywords(remove_keywords)
                        self.assertEqual(sorted(expected_keywords), sorted(track.keywords))
                backend2.scan()
                self.assertEqual(
                    backend2._get_current_keywords(backend2[0]),  # pylint: disable=protected-access
                    backend2[0].keywords)
                self.assertEqual(sorted(expected_keywords), backend2[0].keywords)

    def test_legal_categories(self):
        """Check if our fixed list of categories still matches the online service."""
        with self.temp_backend(Directory) as serverdirectory:
            for cls in (MMT, GPSIES, TrackMMT):
                if cls.is_disabled():
                    continue
                with self.temp_backend(cls, clear_first=False, cleanup=False) as backend:
                    if cls is TrackMMT:
                        with self.lifetrackserver(servername='localhost', port=12398, directory=serverdirectory.url):
                            downloaded = backend._download_legal_categories()  # pylint: disable=protected-access
                            self.assertEqual(sorted(backend.legal_categories), downloaded)
                    else:
                        downloaded = backend._download_legal_categories()  # pylint: disable=protected-access
                        self.assertEqual(sorted(backend.legal_categories), downloaded)



# -*- coding: utf-8 -*-


# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""Tests for gpxity.backends."""

import unittest
import importlib
import pwd
import os
import io
import datetime
import time
import random
from inspect import getmembers, isclass, getmro
from pkgutil import get_data
import tempfile
from contextlib import contextmanager
from subprocess import Popen
import logging

import gpxpy
from gpxpy.gpx import GPXTrackPoint

from ...track import Track
from ...backend import Backend
from ...backends import Mailer
from ...auth import Authenticate
from .. import GPSIES

# pylint: disable=attribute-defined-outside-init

__all__ = ['BasicTest']

from .. import Directory


class BasicTest(unittest.TestCase):

    """define some helpers.

    Attributes:
        all_backend_classes: a list of all backend implementations

    """

    all_backend_classes = None

    def setUp(self):  # noqa
        """define test specific Directory.prefix."""
        Authenticate.path = os.path.join(os.path.dirname(__file__), 'test_auth_cfg')
        self.logger = logging.getLogger()
        self.logger.level = logging.DEBUG
        self.logger.debug('auth file now is %s', Authenticate.path)
        self.start_time = datetime.datetime.now()
        self.unicode_string1 = 'unicode szlig: ß'
        self.unicode_string2 = 'something japanese:の諸問題'
        Directory.prefix = 'gpxity.' + '.'.join(self.id().split('.')[-2:])
        path = tempfile.mkdtemp(prefix=Directory.prefix)

        if not os.path.exists(path):
            os.mkdir(path)

    def tearDown(self):  # noqa
        """Check if there are still /tmp/gpxitytest.* directories."""
        must_be_empty = tempfile.mkdtemp(prefix=Directory.prefix)
        os.rmdir(must_be_empty)
        timedelta = datetime.datetime.now() - self.start_time
        self.logger.debug('%s seconds ', timedelta.seconds)
        logging.shutdown()

    @staticmethod
    def _get_gpx_from_test_file(name: str):
        """get data from a predefined gpx file.

        name is without .gpx

        Returns:=
            A GPX object

        """
        gpx_test_file = os.path.join(os.path.dirname(__file__), '{}.gpx'.format(name))
        if not os.path.exists(gpx_test_file):
            raise Exception('MMTTests needs a GPX file named {}.gpx for testing in {}'.format(
                name, os.getcwd()))
        return gpxpy.parse(io.StringIO(get_data(__package__, '{}.gpx'.format(name)).decode('utf-8')))

    @classmethod
    def create_test_track(
            cls, count: int = 1, idx: int = 0, category: str = None, public: bool = False,
            start_time=None, end_time=None):
        """create a :class:`~gpxity.Track`.

        It starts off with **test.gpx** and appends a
        last track point, it also changes the time stamp of the last point.
        This is done using **count** and **idx**: The last point is set such that
        looking at the tracks, they all go in a different direction clockwise, with an angle
        in degrees of :literal:`360 * idx / count`.

        Args:
            count: See above. Using 1 as default if not given.
            idx: See above. Using 0 as default if not given.
            category: The wanted value for the track.
                Default: if count == len(:attr:`Track.legal_categories <gpxity.Track.legal_categories>`),
                the default value will be legal_categories[idx].
                Otherwise a random value will be applied.
            public: should the tracks be public or private?
            start_time: If given, assign it to the first point and adjust all following times
            end_time: explicit time for the last point. If None: See above.

        Returns:
            (~gpxity.Track): A new track not bound to a backend

        """
        if BasicTest.all_backend_classes is None:
            BasicTest.all_backend_classes = BasicTest._find_backend_classes()
        gpx = cls._get_gpx_from_test_file('test')
        if start_time is not None:
            _ = start_time - gpx.tracks[0].segments[0].points[0].time
            gpx.adjust_time(_)
        last_points = gpx.tracks[-1].segments[-1].points
        if end_time is None:
            end_time = last_points[-1].time + datetime.timedelta(hours=10, seconds=idx)
        new_point = GPXTrackPoint(
            latitude=last_points[-1].latitude, longitude=last_points[-1].longitude + 0.001, time=end_time)
        _ = gpxpy.geo.LocationDelta(distance=1000, angle=360 * idx / count)
        new_point.move(_)
        last_points.append(new_point)

        # now set all times such that they are in order with this track and do not overlap
        # with other test tracks
        _ = gpx.tracks[0].segments[0].points[0].time
        duration = new_point.time - _ + datetime.timedelta(seconds=10)
        for point in gpx.walk(only_points=True):
            point.time += duration * idx

        result = Track(gpx=gpx)
        result.title = 'Random GPX # {}'.format(idx)
        result.description = 'Description to {}'.format(gpx.name)
        if category:
            result.category = category
        elif count == len(Track.legal_categories):
            result.category = Track.legal_categories[idx]
        else:
            result.category = random.choice(Track.legal_categories)
        result.public = public
        return result

    @staticmethod
    def _random_datetime():
        """random datetime between now() - 10 days and now().

        Returns:
            A random datetime

        """
        end = datetime.datetime.now()
        start = end - datetime.timedelta(days=10)
        delta = end - start
        int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
        random_second = random.randrange(int_delta)
        return start + datetime.timedelta(seconds=random_second)

    @staticmethod
    def _random_keywords(count=100):
        """A set of random keywords, but always the same.

        We do not want to generate too many tag ids for MMT.

        Returns:
            A set of random keywords

        """
        state = random.getstate()
        try:
            random.seed(1)
            basis = 'abcdefghijklmnopqrstuvwxyz'
            basis += basis.upper()
            basis += '/-_+.% $"|\\'
            result = set()
            while len(result) < count:
                _ = ''.join(random.choice(basis) for x in range(4))
                result.add(_.strip())
            return result
        finally:
            random.setstate(state)

    @classmethod
    def _random_points(cls, count=100):
        """Get some random points.

        Returns:
            A list with count points

        """
        result = list()
        start_time = cls._random_datetime()
        for _ in range(count):
            point = GPXTrackPoint(
                latitude=random.uniform(0.0, 90.0),
                longitude=random.uniform(0.0, 180.0), elevation=_,
                time=start_time + datetime.timedelta(seconds=10 * _))
            result.append(point)
        return result

    def assertSameTracks(self, backend1, backend2, with_category=True):  # noqa pylint: disable=invalid-name
        """both backends must hold identical tracks."""
        self.maxDiff = None  # pylint: disable=invalid-name
        if backend1 != backend2:
            with_last_time = not (isinstance(backend1, GPSIES) or isinstance(backend2, GPSIES))
            keys1 = sorted(x.key(with_category, with_last_time) for x in backend1)
            keys2 = sorted(x.key(with_category, with_last_time) for x in backend2)
            self.assertEqual(keys1, keys2)

    def assertEqualTracks(self, track1, track2, xml: bool = False, with_category: bool = True):  # noqa pylint: disable=invalid-name
        """both tracks must be identical. We test more than necessary for better test coverage.

        Args:

            xml: if True, also compare to_xml()"""
        self.maxDiff = None

        # GPSIES: when uploading tracks. GPSIES sometimes assigns new times to all points,
        # starting at 2010-01-01 00:00. Until I find the reason, ignore point times for comparison.
        with_last_time = not (isinstance(track1.backend, GPSIES) or isinstance(track2.backend, GPSIES))

        self.assertEqual(track1.key(with_category, with_last_time), track2.key(with_category, with_last_time))
        self.assertTrue(track1.points_equal(track2))
        if xml:
            self.assertEqual(track1.gpx.to_xml(), track2.gpx.to_xml())

    def assertNotEqualTracks(self, track1, track2, with_category: bool = True):  # noqa pylint: disable=invalid-name
        """both tracks must be different. We test more than necessary for better test coverage."""
        self.assertNotEqual(track1.key(with_category), track2.key(with_category))
        self.assertFalse(track1.points_equal(track2))
        self.assertNotEqual(track1.gpx.to_xml(), track2.gpx.to_xml())

    def assertTrackFileContains(self, track, string):  # noqa pylint: disable=invalid-name
        """Assert that string is in the physical file. Works only for Directory backend."""
        with open(track.backend.gpx_path(track.id_in_backend)) as trackfile:
            data = trackfile.read()
        self.assertIn(string, data)

    def setup_backend(  # pylint: disable=too-many-arguments
            self, cls_, username: str = None, url: str = None, count: int = 0,
            cleanup: bool = True, clear_first: bool = True, category: str = None,
            public: bool = False):
        """set up an instance of a backend with count tracks.

        If count == len(:attr:`Track.legal_categories <gpxity.Track.legal_categories>`),
        the list of tracks will always be identical. For an example
        see :meth:`TestBackends.test_all_category <gpxity.backends.test.test_backends.TestBackends.test_all_category>`.

        Args:
            cls_ (Backend): the class of the backend to be created
            username: use this to for a specific accout name. Default is 'gpxitytest'
            url: for the backend
            count: how many random tracks should be inserted?
            cleanup: If True, remove all tracks when done. Passed to the backend.
            clear_first: if True, first remove all existing tracks
            public: should the tracks be public or private?

        Returns:
            the prepared Backend

        """

        result = cls_(url, auth=username or 'gpxitytest', cleanup=cleanup)
        if clear_first:
            result.remove_all()
        if count:
            # if count == 0, skip this. Needed for write-only backends like Mailer.
            while count > len(result):
                track = self.create_test_track(count, len(result), category=category, public=public)
                result.add(track)
            self.assertGreaterEqual(len(result), count)
            if clear_first:
                self.assertEqual(len(result), count)
        return result

    @staticmethod
    @contextmanager
    def lifetrackserver(directory, servername, port):
        """Start and ends a server for lifetrack testing."""
        if Mailer.is_disabled():
            raise Exception('gpxity_server cannot run because Mailer is disabled')
        cmdline = 'gpxity_server --loglevel debug --servername {} --port {} --directory {} --mailto {}'.format(
            servername, port, directory, pwd.getpwuid(os.geteuid()).pw_name)
        process = Popen(cmdline.split())
        try:
            time.sleep(1)  # give the server time to start
            yield
        finally:
            process.kill()

    @contextmanager
    def temp_backend(self, cls_, url=None, count=0,  # pylint: disable=too-many-arguments
                     cleanup=True, clear_first=True, category=None,
                     public: bool = False, username=None):
        """Just like setup_backend but usable as a context manager. which will call destroy() when done."""
        tmp_backend = self.setup_backend(cls_, username, url, count, cleanup, clear_first, category, public)
        try:
            yield tmp_backend
        finally:
            tmp_backend.destroy()

    @staticmethod
    def clone_backend(backend):
        """return a clone of backend."""
        return backend.__class__(backend.url, backend.auth)

    @staticmethod
    def _find_backend_classes(with_skip: bool = False):
        """Find all backend classes. Those will be tested.

        Args:
            with_skip: if True, also finds those with skip_test=False

        Returns:
            A list of backend classes

        """
        backends_directory = __file__
        while not backends_directory.endswith('backends'):
            backends_directory = os.path.dirname(backends_directory)
        if not os.path.exists(backends_directory):
            raise Exception('we are not where we should be')
        result = list()
        mod_names = os.listdir(backends_directory)
        for mod in mod_names:
            if not mod.endswith('.py'):
                continue
            mod = mod.replace('.py', '')
            if mod == '__init__':
                continue
            try:
                imported = importlib.__import__(mod, globals(), locals(), level=2)
                for name, cls in getmembers(imported, isclass):
                    if name in imported.__all__ and Backend in getmro(cls)[1:]:
                        # isinstance and is do not work here
                        if cls.is_disabled():
                            continue
                        if with_skip or not cls.skip_test:
                            result.append(cls)
            except ImportError:
                pass
        # sort because we want things reproducibly
        return sorted(set(result), key=lambda x: x.__class__.__name__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

# PYTHON_ARGCOMPLETE_OK
# for command line argument completion, put this into your .bashrc:
# eval "$(register-python-argcomplete gpxdo)"
# or see https://argcomplete.readthedocs.io/en/latest/


"""lifetrack_client is a command line tool for lifetrack testing, client side."""

import argparse
import os
import sys
import logging
import time

try:
    import argcomplete
    # pylint: disable=unused-import
    from argcomplete import ChoicesCompleter  # noqa
except ImportError:
    pass

from gpxpy import gpx as mod_gpx
GPX = mod_gpx.GPX
GPXTrack = mod_gpx.GPXTrack
GPXTrackSegment = mod_gpx.GPXTrackSegment


# This uses not the installed copy but the development files
_ = os.path.dirname(sys.path[0] or sys.path[1])
if os.path.exists(os.path.join(_, 'gpxity', '__init__.py')):
    sys.path.insert(0, _)
# pylint: disable=wrong-import-position
from gpxity import Track, Lifetrack, Backend, Directory, MMT, GPSIES, TrackMMT  # noqa


class Main:

    """this is where the work is done."""

    def __init__(self):
        """No args."""
        # pylint: disable=too-many-branches,too-many-nested-blocks

        self.exit_code = 0
        self.options = None
        self.parse_commandline()
        if self.exit_code:
            return
        self.logger = logging.getLogger()
        self.logger.setLevel(self.options.loglevel.upper())
        self.source = None
        try:
            source = self.instantiate_object(self.options.source)
            assert isinstance(source, Track)
            backend = self.instantiate_object(self.options.backend)
            assert isinstance(backend, Backend)
            try:
                life = Lifetrack(backend)
                life.set_title('Todays Lifetrack')
                all_points = list(source.points())
                life.update(all_points[:5])
                delay = self.delay()
                for point in all_points[5:]:
                    time.sleep(next(delay))
                    life.update([point])
                time.sleep(next(delay))
                life.end()
            finally:
                backend.destroy()
        except Exception as _:  # pylint: disable=broad-except
            self.error(_)

    @staticmethod
    def delay():
        """Varying delay for testing."""
        while True:
            yield 7
            yield 15

    def error(self, msg, exit_code=None):
        """Print the error message.
        Sets the process exit code.
        With --debug, re-raises the exception."""
        self.logger.error(msg)
        self.exit_code = exit_code or 1
        if self.logger.level == logging.DEBUG:
            raise msg

    def instantiate_object(self, name):
        """return a backend for name.
        If name is a single track, the returned backend has a match filtering
        only this one wanted track."""
        # pylint: disable=too-many-branches
        result = account = track_id = None
        if ':' in name and name.split(':')[0].upper() in ('MMT', 'GPSIES', 'TRACKMMT'):
            clsname = name.split(':')[0].upper()
            rest = name[len(clsname) + 1:]
            if '/' in rest:
                if rest.count('/') > 1:
                    raise Exception('wrong syntax in {}'.format(name))
                account, track_id = rest.split('/')
            else:
                account = rest
            if clsname == 'MMT':
                result = MMT(auth=account, timeout=self.options.timeout)
            elif clsname == 'TRACKMMT':
                result = TrackMMT(auth=account, timeout=self.options.timeout)
            elif clsname == 'GPSIES':
                result = GPSIES(auth=account, timeout=self.options.timeout)
        else:
            if os.path.isdir(name):
                account = name
                result = Directory(url=account)
            else:
                if name.endswith('.gpx'):
                    name = name[:-4]
                if os.path.isfile(name + '.gpx'):
                    account = os.path.dirname(name) or '.'
                    track_id = os.path.basename(name)
                result = Directory(url=account)
        if account is None:
            raise Exception('{} not found'.format(name))
        if track_id:
            result = result[track_id]
        return result

    def parse_commandline(self):
        """into self.options."""
        # pylint: disable=too-many-statements, too-many-branches
        parser = argparse.ArgumentParser('lifetrack_client')
        parser.add_argument('--source', help='the track with test data')
        parser.add_argument('--backend', help='the server')
        parser.add_argument(
            '--loglevel', help='set the loglevel',
            choices=('debug', 'info', 'warning', 'error'), default='error')
        parser.add_argument('--timeout', help="""
            Timeout: Either one value in seconds or two comma separated values: The first one is the connection
            timeout, the second one is the read timeout. Default is to wait forever.""", type=str, default=None)

        try:
            argcomplete.autocomplete(parser)
        except NameError:
            pass

        if len(sys.argv) < 2:
            parser.print_usage()
            sys.exit(2)

        self.options = parser.parse_args()

        if self.options.timeout is not None:
            if ',' in self.options.timeout:
                self.options.timeout = tuple(float(x) for x in self.options.timeout.split(','))
            else:
                self.options.timeout = float(self.options.timeout)


sys.exit(Main().exit_code)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
This implements :class:`gpxity.MMT` for http://www.mapmytracks.com.

There are some problems with the server running at mapmytracks.com:
    * it is not possible to change an existing track - if the track changes, the
      track must be re-uploaded and gets a new track id This invalididates
      references held by other backend instances (maybe on other clients).
      But I could imagine that most similar services have this problem too.
    * does not support GPX very well beyond track data. One problem is that
      it does not support gpx.time, it ignores it in uploads and uses the time
      of the earliest trackpoint. To be consistent, Gpxity follows that for now
      and does not respect gpx.time either.
    * there is an official description of an API at https://github.com/MapMyTracks
      but this does not implement everything needed. For the missing parts we
      simulate what a web browser would do, see :meth:`MMT._read_all` and
      :meth:`MMT._write_attribute`. Of course that could fail if MMT changes its site.
      Which is true for the api itself, it can and does get incompatible changes at
      any time without notice to users or deprecation periods.
    * downloading tracks with that abi is very slow and hangs forever for big tracks
      (at least this was so in Feb 2017, maybe have to test again occasionally).
    * not all parts of MMT data are supported like images (not interesting for me,
      at least not now).

"""

# TODO: logout

from xml.etree import ElementTree
import html
from html.parser import HTMLParser
import datetime
from collections import defaultdict
import requests

from .. import Backend, Track
from ..version import VERSION


__all__ = ['MMT']


def _convert_time(raw_time) ->datetime.datetime:
    """MMT uses Linux timestamps. Converts that into datetime.

    Args:
        raw_time (int): The linux timestamp from the MMT server

    Returns:
        The datetime

    """
    return datetime.datetime.utcfromtimestamp(float(raw_time))


class ParseMMTCategories(HTMLParser):  # pylint: disable=abstract-method

    """Parse the legal values for category from html."""

    def __init__(self):
        """See class docstring."""
        super(ParseMMTCategories, self).__init__()
        self.seeing_category = False
        self.result = ['Cycling']  # The default value

    def handle_starttag(self, tag, attrs):
        """starttag from the parser."""
        # pylint: disable=too-many-branches
        attributes = dict(attrs)
        self.seeing_category = (
            tag == 'input' and 'name' in attributes and attributes['name'].startswith('add-activity'))

    def handle_data(self, data):
        """handle the data."""
        if self.seeing_category:
            _ = data.strip()
            if _ not in self.result:
                self.result.append(_)
            self.seeing_category = False


class ParseMMTTrack(HTMLParser):  # pylint: disable=abstract-method

    """get some attributes available only on the web page.

    Of course, this is highly unreliable. Just use what we can get."""

    result = dict()

    def __init__(self, backend):
        """See class docstring."""
        super(ParseMMTTrack, self).__init__()
        self.backend = backend
        self.seeing_category = False
        self.seeing_title = False
        self.seeing_description = False
        self.seeing_status = False
        self.seeing_tag = None
        self.result['mid'] = None
        self.result['title'] = None
        self.result['description'] = None
        self.result['category'] = None
        self.result['category_from_title'] = None
        self.result['category_3'] = None
        self.result['public'] = None
        self.result['tags'] = dict()  # key: name, value: id

    def handle_starttag(self, tag, attrs):
        """starttag from the parser."""
        # pylint: disable=too-many-branches
        self.seeing_title = False
        self.seeing_description = False
        self.seeing_category = False
        self.seeing_status = False
        self.seeing_tag = None
        attributes = defaultdict(str)
        for key, value in attrs:
            attributes[key] = value
        if tag == 'input':
            value = attributes['value'].strip()
            if (attributes['id'] == 'activity_type' and attributes['type'] == 'hidden'
                    and attributes['name'] == 'activity_type' and value):  # noqa
                self.result['category_3'] = value
            elif (attributes['id'] == 'mid' and attributes['type'] == 'hidden'
                  and attributes['name'] == 'mid'and value):  # noqa
                self.result['mid'] = value
        elif tag == 'div' and attributes['class'] == 'panel' and 'data-activity' in attributes:
            self.result['category'] = attributes['data-activity']
        elif tag == 'span' and attributes['class'] == 'privacy-status':
            self.seeing_status = True
        elif tag == 'title':
            self.seeing_category = True
        elif tag == 'h2' and attributes['id'] == 'track-title':
            self.seeing_title = True
        elif tag == 'p' and attributes['id'] == 'track-desc':
            self.seeing_description = True
        elif tag == 'a' and attributes['class'] == 'tag-link' and attributes['rel'] == 'tag':
            assert attributes['id'].startswith('tag-')
            self.seeing_tag = attributes['id'].split('-')[2]

    def handle_data(self, data):
        """data from the parser."""
        if not data.strip():
            return
        if self.seeing_title:
            self.result['title'] = data.strip()
        if self.seeing_description:
            self.result['description'] = data.strip()
        if self.seeing_category:
            try:
                _ = data.split('|')[1].split('@')[0].strip()
                self.result['category_from_title'] = ' '.join(_.split(' ')[:-2])
            except BaseException:
                self.backend.logger.warning('%s: Cannot parse %s', self.backend.identifier(), data)
                self.result['category_from_title'] = ''
        if self.seeing_status:
            self.result['public'] = data.strip() != 'Only you can see this activity'
        if self.seeing_tag:
            self.result['tags'][data.strip()] = self.seeing_tag


class MMTRawTrack:

    """raw data from mapmytracks.get_tracks."""

    # pylint: disable=too-few-public-methods
    def __init__(self, xml):
        """See class docstring."""
        self.track_id = xml.find('id').text
        self.title = html.unescape(xml.find('title').text)
        self.time = _convert_time(xml.find('date').text)
        self.category = html.unescape(xml.find('activity_type').text)
        self.distance = float(xml.find('distance').text)


class MMT(Backend):

    """The implementation for MapMyTracks.

    The track ident is the number given by MapMyTracks.

    MMT knows tags. We map :attr:`Track.keywords <gpxity.Track.keywords>` to MMT tags. MMT will
    change keywords: It converts the first character to upper case. See
    :attr:`Track.keywords <gpxity.Track.keywords>` for how Gpxity handles this.

    Args:
        url (str): The Url of the server. Default is http://mapmytracks.com
        auth (tuple(str, str)): Username and password
        cleanup (bool): If True, :meth:`~gpxity.Backend.destroy` will remove all tracks in the
            user account.
         timeout: If None, there are no timeouts: Gpxity waits forever. For legal values
            see http://docs.python-requests.org/en/master/user/advanced/#timeouts

    """

    # pylint: disable=abstract-method

    # skip_test = True

    _default_description = 'None yet. Let everyone know how you got on.'

    legal_categories = (
        'Cycling', 'Running', 'Mountain biking', 'Sailing', 'Walking', 'Hiking',
        'Driving', 'Off road driving', 'Motor racing', 'Motorcycling', 'Enduro',
        'Skiing', 'Cross country skiing', 'Canoeing', 'Kayaking', 'Sea kayaking',
        'SUP boarding', 'Rowing', 'Swimming', 'Windsurfing', 'Orienteering',
        'Mountaineering', 'Skating', 'Horse riding', 'Hang gliding', 'Hand cycling',
        'Gliding', 'Flying', 'Kiteboarding', 'Snowboarding', 'Paragliding',
        'Hot air ballooning', 'Nordic walking', 'Miscellaneous', 'Skateboarding',
        'Snowshoeing', 'Jet skiing', 'Powerboating', 'Wheelchair', 'Indoor cycling')

    _category_encoding = {
        'Cabriolet': 'Driving',
        'Coach': 'Miscellaneous',
        'Crossskating': 'Skating',
        'Handcycle': 'Cycling',
        'Motorhome': 'Driving',
        'Pack animal trekking': 'Hiking',
        'Pedelec': 'Cycling',
        'Stand up paddle boarding': 'SUP boarding',
        'Train': 'Miscellaneous',
    }

    default_url = 'http://www.mapmytracks.com'

    # MMT only accepts one simultaneous lifetracker per login. We make sure
    # that at least this process does not try to run several at once.
    # This check is now too strict: We forbid multiple lifetrackers even if
    # every MMT account only gets one.
    _current_lifetrack = None

    def __init__(self, url=None, auth=None, cleanup=False, timeout=None):
        """See class docstring."""
        if url is None:
            url = self.default_url
        super(MMT, self).__init__(url, auth, cleanup, timeout)
        self.__mid = -1  # member id at MMT for auth
        self.__is_free_account = None
        self.__tag_ids = dict()  # key: tag name, value: tag id in MMT. It seems that MMT
        # has a lookup table and never deletes there. So a given tag will always get
        # the same ID. We use this fact.
        # MMT internally capitalizes tags but displays them lowercase.
        self._last_response = None  # only used for debugging
        self.https_url = self.url.replace('http:', 'https:')

    def _download_legal_categories(self):
        """Needed only for unittest.

        Returns: list(str)
            all legal values for category.

        """
        response = requests.get(self.url + '/explore/wall', timeout=self.timeout)
        category_parser = ParseMMTCategories()
        category_parser.feed(response.text)
        return sorted(category_parser.result)

    @property
    def session(self):
        """The requests.Session for this backend. Only initialized once.

        Returns:
                The session

        """
        ident = self.identifier()
        if ident not in self._session:
            if not self.auth:
                raise self.BackendException('{}: Needs authentication data'.format(self.url))
            self._session[ident] = requests.Session()
            # I have no idea what ACT=9 does but it seems to be needed
            payload = {'username': self.auth[0], 'password': self.auth[1], 'ACT': '9'}
            login_url = '{}/login'.format(self.https_url)
            response = self._session[ident].post(
                login_url, data=payload, timeout=self.timeout)
            if 'You are now logged in.' not in response.text:
                raise self.BackendException('Login as {} failed'.format(self.auth[0]))
            cookies = requests.utils.dict_from_cookiejar(self._session[ident].cookies)
            self._session[ident].cookies = requests.utils.cookiejar_from_dict(cookies)
        return self._session[ident]

    def decode_category(self, value: str) ->str:
        """Translate the value from MMT into internal one.

        Since gpxity once decided to use MMT definitions for tracks, this should mostly be 1:1 here.

        Returns:
            the decoded value

        """
        if value not in Track.legal_categories:
            raise self.BackendException('MMT gave us an unknown category={}'.format(value))
        return value

    def encode_category(self, value: str) ->str:
        """Translate internal value into MMT value.

        Returns:
            the translated value

        """
        if value in self.legal_categories:
            return value
        if value not in self._category_encoding:
            raise self.BackendException('MMT has no equivalent for {}'.format(value))
        return self._category_encoding[value]

    @property
    def mid(self):
        """the member id on MMT belonging to auth.

        Returns:
            The mid

        """
        if self.__mid == -1:
            self._parse_homepage()
        return self.__mid

    @property
    def is_free_account(self):
        """Return True if the current account is not PLUS enabled."""
        if self.__is_free_account is None:
            self._parse_homepage()
        return self.__is_free_account

    def _parse_homepage(self):
        """Get some interesting values from the home page."""
        response = self.session.get(self.url)
        self.__is_free_account = 'href="/plus">Upgrade to PLUS' in response.text
        page_parser = ParseMMTTrack(self)
        page_parser.feed(response.text)
        self.__mid = page_parser.result['mid']
        self.__tag_ids.update(page_parser.result['tags'])
        self._check_tag_ids()

    @staticmethod
    def _encode_keyword(value):
        """mimic the changes MMT applies to tags.

        Returns:
            The changed keywords

        """
        return ' '.join(x.capitalize() for x in value.split())

    def _check_tag_ids(self):
        """Assert that all tags conform to what MMT likes."""
        for _ in self.__tag_ids:
            assert _[0].upper() == _[0], self.__tag_ids

    def _found_tag_id(self, tag, id_):
        """We just learned about a new tag id. They never change for a given string."""
        self.__tag_ids[tag] = id_
        self._check_tag_ids()

    def __post(
            self, with_session: bool = False, url: str = None, data: str = None, expect: str = None, **kwargs) ->str:
        """Helper for the real function with some error handling.

        Args:
            with_session: If given, use self.session. Otherwise, use basic auth.
            url:  Will be appended to self.url. Default is api/. For the basic url, pass an empty  string.
            data: should be xml and will be encoded. May be None.
            expect: If given, raise an error if this string is not part of the server answer.
            kwargs: a dict for post(). May be None. data and kwargs must not both be passed.

        Returns:
            the result

        """
        if url is None:
            url = 'api/'
        full_url = self.url + url
        headers = {'DNT': '1'}  # do not track
        if data:
            data = data.encode('ascii', 'xmlcharrefreplace')
        else:
            data = kwargs
        try:
            if with_session:
                response = self.session.post(
                    full_url, data=data, headers=headers, timeout=self.timeout)
            else:
                response = requests.post(
                    full_url, data=data, headers=headers, auth=self.auth, timeout=self.timeout)
        except requests.exceptions.ReadTimeout:
            self.logger.error('%s: timeout for %s', self.identifier(), data)
            raise
        self._last_response = response  # for debugging
        if response.status_code != requests.codes.ok:  # pylint: disable=no-member
            self.__handle_post_error(full_url, data, response)
            return None
        result = response.text
        if (result == 'access denied') or (expect and expect not in result):
            raise self.BackendException('{}: expected {} in {}'.format(data, expect, result))
        if result.startswith('<?xml'):
            try:
                result = ElementTree.fromstring(result)
            except ElementTree.ParseError:
                raise self.BackendException('POST {} has parse error: {}'.format(data, response.text))
            result_type = result.find('type')
            if result_type is not None and result_type.text == 'error':
                reason = result.find('reason').text if result.find('reason') else 'no reason given'
                raise self.BackendException('{}: {}'.format(data, reason))
        return result

    @classmethod
    def __handle_post_error(cls, url, data, result):
        """we got status_code != ok."""
        try:
            result.raise_for_status()
        except BaseException as exc:
            if isinstance(data, str) and 'request' in data:
                _ = data['request']
            else:
                _ = data
            raise cls.BackendException('{}: {} {} {}'.format(exc, url, _, result.text))

    def _write_attribute(self, track, attribute):
        """change an attribute directly on mapmytracks.

        Note that we specify iso-8859-1 but use utf-8. If we correctly specify utf-8 in

        the xml encoding, mapmytracks.com aborts our connection."""
        attr_value = getattr(track, attribute)
        if attribute == 'description' and attr_value == self._default_description:
            attr_value = ''
        # MMT returns 500 Server Error if we set the title to an empty string
        if attribute == 'title' and not attr_value:
            attr_value = 'no title'
        data = '<?xml version="1.0" encoding="ISO-8859-1"?>' \
            '<message><nature>update_{attr}</nature><eid>{eid}</eid>' \
            '<usr>{usrid}</usr><uid>{uid}</uid>' \
            '<{attr}>{value}</{attr}></message>'.format(
                attr=attribute,
                eid=track.id_in_backend,
                usrid=self.auth[0],
                value=attr_value,
                uid=self.session.cookies['exp_uniqueid'])
        self.__post(with_session=True, url='assets/php/interface.php', data=data, expect='success')

    def _write_title(self, track):
        """change title on remote server."""
        self._write_attribute(track, 'title')

    def _write_description(self, track):
        """change description on remote server."""
        self._write_attribute(track, 'description')

    def _write_public(self, track):
        """change public/private on remote server."""
        self.__post(
            with_session=True, url='user-embeds/statuschange-track', expect='access granted',
            mid=self.mid, tid=track.id_in_backend,
            hash=self.session.cookies['exp_uniqueid'],
            status=1 if track.public else 2)
        # what a strange answer

    def _write_category(self, track):
        """change category directly on mapmytracks.

        Note that we specify iso-8859-1 but use utf-8. If we correctly specify utf-8 in

        the xml encoding, mapmytracks.com aborts our connection."""
        self.__post(
            with_session=True, url='handler/change_activity', expect='ok',
            eid=track.id_in_backend, activity=self.encode_category(track.category))

    def _current_tags(self, track):
        """Return all current MMT tags.

        Returns:

            A sorted unique list"""
        page_scan = self._scan_track_page(track)
        return list(sorted(set(page_scan['tags'])))

    def _write_add_keywords(self, track, values):
        """Add keyword as MMT tag.

        MMT allows adding several at once, comma separated,

        and we allow this too. But do not expect this to work with all backends."""
        if not values:
            return
        values = ','.join(sorted(values.split(',')))
        data = '<?xml version="1.0" encoding="ISO-8859-1"?>' \
            '<message><nature>add_tag</nature><eid>{eid}</eid>' \
            '<usr>{usrid}</usr><uid>{uid}</uid>' \
            '<tagnames>{value}</tagnames></message>'.format(
                eid=track.id_in_backend,
                usrid=self.auth[0],
                value=values,
                uid=self.session.cookies['exp_uniqueid'])
        text = self.__post(with_session=True, url='assets/php/interface.php', data=data, expect='success')
        values = [x.strip() for x in values.split(',')]
        ids = (text.find('ids').text or '').split(',')
        tags = (text.find('tags').text or '').split(',')
        if values != tags or len(ids) != len(values):
            if values != tags:
                raise self.BackendException(
                    '{}: _write_add_keywords({}): MMT does not like some of your keywords: mmt tags={}'.format(
                        self._current_track, ','.join(values), ','.join(tags)))
            if len(ids) != len(values):
                raise self.BackendException(
                    '{}: _write_add_keywords({}): MMT does not like some of your keywords: mmt ids={}'.format(
                        self._current_track, ','.join(values), ','.join(ids)))
        for tag, id_ in zip(values, ids):
            self._found_tag_id(tag, id_)

    def _write_remove_keywords(self, track, values):
        """Remove keywords from track."""
        # with Track.batch_changes() active, track.keywords is already in the future
        # state after all batched changes have been applied, but we need the current
        # state. Ask MMT.
        current = self._get_current_keywords(track)
        wanted = set(current) - {x.strip() for x in values.split(',')}
        if True:  # pylint: disable=using-constant-test
            # First remove all keywords and then re-add the still wanted ones. This works!
            # Because even if MMT does not remove the correct keyword, it always does
            # remove one of them.
            for value in current:
                self._remove_single_keyword(track, value)
            self._write_add_keywords(track, ','.join(wanted))
        else:
            # Specifically remove unwanted keywords. This does not work, MMT does not
            # always remove the correct keyword. No idea why.
            for value in values.split(','):
                if value in current:
                    self._remove_single_keyword(track, value)

    def _remove_single_keyword(self, track, value):
        """Remove a specific keyword from track. Does not work correctly, see above."""
        tag = value.strip()
        if tag not in self.__tag_ids:
            self.__tag_ids.update(self._scan_track_page(track)['tags'])
            self._check_tag_ids()
            if tag not in self.__tag_ids:
                raise self.BackendException(
                    '{}: Cannot remove tag {}, MMT does not have an id'.format(self._current_track, tag))
        if tag in self.__tag_ids:
            self.__post(
                with_session=True, url='handler/delete-tag.php',
                tag_id=self.__tag_ids[tag], entry_id=track.id_in_backend)

    def get_time(self) ->datetime.datetime:
        """get MMT server time.

        Returns:
            The server time

        """
        return _convert_time(self.__post(request='get_time').find('server_time').text)

    def _yield_tracks(self):
        """get all tracks for this user."""

        while True:
            old_len = self.real_len()
            response = self.__post(
                request='get_activities', author=self.auth[0],
                offset=old_len)
            chunk = response.find('activities')
            if not chunk:
                return
            for _ in chunk:
                raw_data = MMTRawTrack(_)
                track = self._found_track(raw_data.track_id)
                # pylint: disable=protected-access
                track._header_data['title'] = raw_data.title
                track._header_data['category'] = self.decode_category(raw_data.category)
                track._header_data['time'] = raw_data.time
                track._header_data['distance'] = raw_data.distance
                yield track
            assert self.real_len() > old_len

    def _scan_track_page(self, track):
        """The MMT api does not deliver all attributes we want.
        This gets some more by scanning the web page and
        returns it in page_parser.result"""
        response = self.session.get('{}/explore/activity/{}'.format(
            self.url, track.id_in_backend))
        page_parser = ParseMMTTrack(self)
        page_parser.feed(response.text)
        return page_parser.result

    def _get_current_keywords(self, track):
        """Ask MMT for current keywords, return them as a list."""
        page_scan = self._scan_track_page(track)
        if page_scan['tags']:
            return sorted(page_scan['tags'].keys())
        return list()

    def _use_webpage_results(self, track):
        """Get things directly.

        if the title has not been set, get_activities says something like "Track 2016-09-04 ..."
            while the home page says "Cycling activity". We prefer the value from the home page
            and silently ignore this inconsistency.

         """
        page_scan = self._scan_track_page(track)
        if page_scan['title']:
            track.title = page_scan['title']
        if page_scan['description']:
            _ = page_scan['description']
            if _ == self._default_description:
                _ = ''
            track.description = _
        if page_scan['tags']:
            track.keywords = page_scan['tags'].keys()
        # MMT sends different values of the current track type, hopefully category_3 is always the
        # correct one.
        if page_scan['category_3']:
            track.category = self.decode_category(page_scan['category_3'])
        if page_scan['public'] is not None:
            track.public = page_scan['public']

    def _read_all(self, track):
        """get the entire track."""
        session = self.session
        if session is None:
            # https access not implemented for TrackMMT
            return
        response = session.get('{}/assets/php/gpx.php?tid={}&mid={}&uid={}'.format(
            self.url, track.id_in_backend, self.mid, self.session.cookies['exp_uniqueid']))
        # some tracks download only a few points if mid/uid are not given, but I
        # have not been able to write a unittest triggering that ...
        track.parse(response.text)
        # but this does not give us track type and other things,
        # get them from the web page.
        self._use_webpage_results(track)

    def _remove_ident(self, ident: str):
        """remove on the server."""
        self.__post(
            with_session=True, url='handler/delete_track', expect='access granted',
            tid=ident, hash=self.session.cookies['exp_uniqueid'])

    def _write_all(self, track) ->str:
        """save full gpx track on the MMT server.

        We must upload the title separately.

        Returns:
            The new id_in_backend

        """
        self._current_track = track
        if not track.gpx.get_track_points_no():
            raise self.BackendException('MMT does not accept a track without trackpoints:{}'.format(track))
        response = self.__post(
            request='upload_activity', gpx_file=track.to_xml(),
            status='public' if track.public else 'private',
            description=track.description, activity=self.encode_category(track.category))
        new_ident = response.find('id').text
        if not new_ident:
            raise self.BackendException('No id found in response')
        old_ident = track.id_in_backend
        track.id_in_backend = new_ident
        # the caller will do the above too, never mind
        if 'write_title' in self.supported:
            self._write_title(track)
        # MMT can add several keywords at once
        if track.keywords and 'write_add_keywords' in self.supported:
            self._write_add_keywords(track, ', '.join(track.keywords))
        if old_ident:
            self._remove_ident(old_ident)
        track.id_in_backend = new_ident
        self.logger.debug('%s fully written', track.identifier())
        return new_ident

    @staticmethod
    def __formatted_lifetrack_points(points) ->str:
        """format points for life tracking.

        Returns:
            The formatted points

        """
        _ = list()
        for point in points:
            _.append('{} {} {} {}'.format(
                point.latitude,
                point.longitude,
                point.elevation if point.elevation is not None else 0,
                point.time.timestamp()))
        return ' '.join(_)

    def _lifetrack_start(self, track, points) ->str:
        """Start a new lifetrack with initial points.

        Returns:
            new_ident: New track id

        """
        if self.is_free_account:
            raise Exception('Your free MMT account does not allow lifetracking')
        if MMT._current_lifetrack is not None:
            raise Exception('start: MMT only accepts one simultaneous lifetracker per username')
        MMT._current_lifetrack = track
        result = self.__post(
            url='/',
            request='start_activity',
            title=track.title,
            privacy='public' if track.public else 'private',
            activity=self.encode_category(track.category),
            points=self.__formatted_lifetrack_points(points),
            source='Gpxity',
            version=VERSION,
            expect='activity_started',
            # tags='TODO',
            unique_token='{}'.format(id(track)))
        result = result.find('activity_id').text
        self.logger.error('%s: lifetracking started', self.identifier())
        return result

    def _lifetrack_update(self, track, points):
        """Update a lifetrack with poings.

        Args:
            track: The lifetrack
            points: The new point

        """
        if MMT._current_lifetrack != track:
            raise Exception('update: MMT only accepts one simultaneous lifetracker per username')
        self.__post(
            url='/',
            request='update_activity', activity_id=track.id_in_backend,
            points=self.__formatted_lifetrack_points(points),
            expect='activity_updated')

    def _lifetrack_end(self, track):
        """End a lifetrack.

        Args:
            track: The lifetrack

        """
        if MMT._current_lifetrack != track:
            raise Exception('end: MMT only accepts one simultaneous lifetracker per username')
        self.__post(url='/', request='stop_activity')
        MMT._current_lifetrack = None

    def destroy(self):
        """also close session."""
        super(MMT, self).destroy()
        ident = self.identifier()
        if ident in self._session:
            self._session[ident].close()
